/* -*- Mode:C++ -*- */
/*
 * Copyright (c) Multimedia Communications Lab, Boston University
 *
 * Simulation Program to Evaluate Service Aggregation in VOD
 * =========================================================
 *
 * This work was supported by NSF under grant No. NCR-9523958
 * 
 * DISCLAIMER:
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL
 * MULTIMEDIA COMMUNICATIONS LAB, BOSTON UNIVERSITY OR THE CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 * Contributors:
 *
 * Rajesh Krishnan, 
 * Prithwish Basu,
 * Wang Ke
 */

#ifndef __VOD_CONFIG_HPP__
#define __VOD_CONFIG_HPP__

#include <string.h>

#define FRAME_RATE 30 // 30 frames per sec video 

enum CostForm { BIT, TIME_L_INF, TIME_L_FIN};

typedef struct 
{
  int VERBOSE;	 	// Verbose error messages if set
  int MAXSIMCLOCK; 	// Simulation Run time in secs
  double SIMGRAIN;	// Simulation Granularity 
  double ZIPF_PARAM;	// Parameter for Zipf Distribution 
  char MOVIE_DB[100];	// Name of Movie Database File
  int TOT_USERS;		// Number of Subscribers
  double MV_RQ_RATE;	// Rate of Movie Requests per sec
  double QT_RQ_RATE; 	// Rate of Quit Requests per sec
  double FAST_PLAY_RATE;	// Rate of Fast Play
  int FF_RATE;		// Rate of Fastforward
  int RW_RATE;		// Rate of Rewind
  double BATCH_INTERVAL;	// Batching Interval
  double RNG_RQ_RATE;	// Renege Rate
  double FF_RQ_RATE;	// Rate of Fast Forward Requests per sec
  double FF_DUR;		// Mean Duration of Fast Forward
  double RW_RQ_RATE;	// Rate of Rewind Requests per sec
  double RW_DUR;		// Mean Duration of Rewind 
  double PAU_RQ_RATE;	// Rate of Pause Requests per sec
  double PAU_DUR;		// Mean Pause Duration 
  double CLUST_RECOMP;	// Cluster Recomputation Interval
  int AGG_POLICY;		// Aggregation Policy
  int SEPARATE_IN_CLUSTERS; // to control if the arriving clients should
  // be clustered based on which recomputation interval they arrived

} CFGDATA;

extern int read_config(char *fname, CFGDATA *d);
extern char *strip_lwsc(char *inp);

class Config
{
public:
  CFGDATA data;
  static double user_id;
  static double reset_time;
  static double debug_at_recomp;
  static double stream_id;
  static double snd_stream;
  static int join_info;
  static int allow_recomp;
  static int check_eomv;
  static int plot_time_pos;
  static int stt_streams;
  static int stt_random;
  static int stt_demand;
  static int aggregate;
  static double clust_window;
  static double min_ad_time;
  static double max_ad_time; 
  static double min_video_time;
  static double long_term_ad_time; // T'
  static double long_term_window;  // T
  static int ratio_max_min_ad;
  static int ratio_vmin_amin;
  static int ratio_tprime_amax;
  static double long_term_video;
  static int allowing_interactions;
  static double wait_distr_power;
  static double wait_max_time;
  static double perc_OD;
  static int max_num_chan;
  static int return_ondemand;
  static int recomp_at_event;
  static CostForm cost_formulation;
  static double cost_rate;
  static int cost_mvlen_multiplier;
  static int show_users_per_stream;
  static unsigned short int seed_a0;
  static unsigned short int seed_a1;
  static unsigned short int seed_a2;
  static unsigned int predict;
  static int debug_predict;
  static double debug_predict_value;

  static unsigned int MAXHISTORY;
  Config()
    {
      // Specify defaults
      data.MAXSIMCLOCK = 14400; 
      data.SIMGRAIN = 1.0;
      data.ZIPF_PARAM = 1.0;
      strcpy(data.MOVIE_DB, "mvdb");
      data.TOT_USERS = 100;
      data.MV_RQ_RATE = 1;
      data.QT_RQ_RATE = 0;
      data.FF_RATE = 5;
      data.RW_RATE = 5;
      data.BATCH_INTERVAL = 30;
      data.SEPARATE_IN_CLUSTERS = 0;
    }
  ~Config() {};
  int Setup(char *fname) {return(read_config(fname, &data));};
  void Dumpdata(void);
};

#endif //__VOD_CONFIG_HPP__


