/* -*- Mode:C++ -*- */
/*
 * Copyright (c) Multimedia Communications Lab, Boston University
 *
 * Simulation Program to Evaluate Service Aggregation in VOD
 * =========================================================
 *
 * This work was supported by NSF under grant No. NCR-9523958
 * 
 * DISCLAIMER:
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL
 * MULTIMEDIA COMMUNICATIONS LAB, BOSTON UNIVERSITY OR THE CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 * Contributors:
 *
 * Rajesh Krishnan, 
 * Prithwish Basu,
 * Wang Ke
 */

#ifndef __EVENT_HPP__
#define __EVENT_HPP__

#include "movie.hpp"
#include "subscriber.hpp"

class Event
{
protected:
  double event_time;
  double rate; 
public:
  Event() {};
  Event(double r) {event_time = 0.0; rate = r; GenerateEvent();};
  virtual ~Event() {};
  // Uniform Distributed Inter-Event times
  virtual void GenerateEvent(void) {event_time += 1/rate;};
  virtual void Dumpdata(void) {cout << event_time << "\n";};
  double EventTime() {return(event_time);};
};


class MovieRequest: public Event
{
private:
  Movie *mvchoice;
  Subscriber *requester;
  Movielist *ml;
  Subscriberlist *sl;
  
public:
  MovieRequest();
  MovieRequest(double r, Movielist *m, Subscriberlist *s);
  void ResetParameters(double, Movielist *, Subscriberlist *);
  ~MovieRequest() {};
  virtual void GenerateEvent(void);
  virtual void Dumpdata(void);
  Movie *Mvchoice(void) {return mvchoice;};
  Subscriber *Requester(void) {return requester;};
  double getNextEventTime() {return event_time;};
  double MvLen(void) { return mvchoice->Duration();};
};

class VCRActionRequest: public Event
{
private:
  Subscriber *requester;
  Subscriberlist *sl;
  Request rtype;
  double meanduration;
  
public:
  VCRActionRequest(double r, double mdur, Request rq, Subscriberlist *s); 
  ~VCRActionRequest() {};
  virtual void GenerateEvent(void);
  virtual void Dumpdata(void);
  Subscriber *Requester(void) {return requester;};
  Request RequestType(void) {return rtype;};
  void pType();
  static  void pType(Request);
  void GenerateEventTime();
  void GenerateEventProcess();
};

#endif //__EVENT_HPP__






