/* -*- Mode:C++ -*- */
/*
 * Copyright (c) Multimedia Communications Lab, Boston University
 *
 * Simulation Program to Evaluate Service Aggregation in VOD
 * =========================================================
 *
 * This work was supported by NSF under grant No. NCR-9523958
 * 
 * DISCLAIMER:
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL
 * MULTIMEDIA COMMUNICATIONS LAB, BOSTON UNIVERSITY OR THE CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 * Contributors:
 *
 * Prithwish Basu
 */

#ifndef __INTERFACE_H__
#define __INTERFACE_H__

/* The Clustering Policies */
typedef enum __Policy {EMCLRSMA, EMCLAFL, GREEDY, EMCLRSMAHEUR, RSMA, RSMAHEUR, ADINSPOL} Policy;


/* The status of a stream */
typedef enum __NStatus {NORMAL, ACCEL, INVALID, NSADINS, NSATEOMV} NStatus;


/* The Stream attributes */
typedef struct _stream
{
    int      movie_id;       
    double   stream_id;
    double   position;  
    NStatus   status;
    double   ad_to_go;
    double   time_to_go;
    double   video_to_go;
    double   movie_length;
    double   merging_with;
} STREAM;


/*************************************************************************/
/* Function prototypes needed for interfacing with the simulation driver */
/*************************************************************************/

/* Returns the status of a particular stream at the current point in the simulation*/
/* time -- the current absolute time (in seconds) */
/* position -- the current position in the movie (in seconds) */
extern NStatus get_stream_status(double stream_id, double time, double position, Policy policy);

/* Sets the status of a particular stream in the clustering side */
/* (For interactive modes primarily) */
/* extern void set_stream_status(double stream_id, double position, NStatus status); */

/* Basically a despatcher function */
/* window -- the catchup window (in seconds) */
/* absolute_time -- the absolute sim clock (in seconds) */
extern void invoke_policy(Policy policy, STREAM *stream, int num_streams, 
		   double window, double absolute_time);

#endif /* ! __INTERFACE_H__ */
