/* -*- Mode:C++ -*- */
/*
 * Copyright (c) Multimedia Communications Lab, Boston University
 *
 * Simulation Program to Evaluate Service Aggregation in VOD
 * =========================================================
 *
 * This work was supported by NSF under grant No. NCR-9523958
 * 
 * DISCLAIMER:
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL
 * MULTIMEDIA COMMUNICATIONS LAB, BOSTON UNIVERSITY OR THE CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 * Contributors:
 *
 * Rajesh Krishnan
 */

#ifndef __LIST_HPP__
#define __LIST_HPP__

#include <stdlib.h>

#ifndef SUCCESS
#define SUCCESS 0
#endif

#ifndef FAILURE
#define FAILURE -1
#endif

#define FORWARD 1
#define BACKWARD 2

class Node
{
  friend class List;
#ifdef AD_INSERTION
  friend class AdStrList;
#endif
protected:
  Node *next;
  Node *prev;
  List *mylist;
public:
  static int tot_num_nodes;
  Node() {next = prev = NULL; mylist = NULL;tot_num_nodes++;};
  virtual ~Node() {tot_num_nodes--;};
  virtual Node *Getnext() {return next;};
  virtual Node *Getprev() {return prev;};
  virtual void Dumpdata(void) {};
};

class List
{
protected:
  Node *head;
  Node *tail;
  int num_nodes;
  
public:
  static int tot_num_lists;
  char ptrlist;

  List() {head = tail = NULL; num_nodes = 0;tot_num_lists++; ptrlist = 0;};
  ~List();
  int Numnodes() {return num_nodes;};
  Node *Seek(Node *datum, int steps, char direction);
  Node *Gethead() {return head;};
  Node *Gettail() {return tail;};
  
  Node *Detach(Node *nodptr);
  void Remove(Node *nodptr) {if(nodptr) delete Detach(nodptr);};
  void Prepend(Node *nodptr, Node *before);
  void Append(Node *nodptr, Node *after);
  void Catenate(List *lst);
  
  int Swap(Node *node1, Node *node2);
  void Dumpdata(void);
  void DeleteAllNodes();
};

#endif //__LIST_HPP__
