/* -*- Mode:C++ -*- */
/*
 * Copyright (c) Multimedia Communications Lab, Boston University
 *
 * Simulation Program to Evaluate Service Aggregation in VOD
 * =========================================================
 *
 * This work was supported by NSF under grant No. NCR-9523958
 * 
 * DISCLAIMER:
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL
 * MULTIMEDIA COMMUNICATIONS LAB, BOSTON UNIVERSITY OR THE CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 * Contributors:
 *
 * Prithwish Basu
 */

#ifndef __RSMA_H__
#define __RSMA_H__
#include "stream.h"

/* This denotes a point in the RSA grid */
typedef struct _point 
{
    double    x;
    double    y;
} POINT;

/* This denotes a graph */
typedef struct _graph
{
    int    num;
    POINT  *arr;
} GRAPH;


/* RSMA stuff */
POINT    parent(POINT, POINT);
POINT    *make_graph (CLUSTER , STREAM *, int *);
int      int_calculate(int i, int j, int n, int *array);
void     int_assign(int i, int j, int n, int *array, int val);
double   dbl_calculate(int i, int j, int n, double *array);
void     dbl_assign(int i, int j, int n, double *array, double val);

void     print_graph(POINT *, int);
void     output_streams(STREAM *, int);
STREAM   get_rsma_state(double, STREAM *, double, CLUSTER *, int, GRAPH *, int **);
void     compute_rsma(CLUSTER, STREAM *, int **);
void     call_rsma(STREAM *streams, CLUSTER *clusters, GRAPH *graph, int **rsma, double W);
void     call_rsma_only(STREAM *streams, CLUSTER *clusters, GRAPH *graph, int **rsma);
void     call_rsma_heur(STREAM *streams, CLUSTER *clusters, GRAPH *graph, int **rsma, double W);
void     call_rsma_heur_only(STREAM *streams, CLUSTER *clusters, GRAPH *graph, int **rsma);
void     er_alloc();
void     er_free();
void     r_alloc();
void     r_free();

#endif
