/* -*- Mode:C++ -*- */
/*
 * Copyright (c) Multimedia Communications Lab, Boston University
 *
 * Simulation Program to Evaluate Service Aggregation in VOD
 * =========================================================
 *
 * This work was supported by NSF under grant No. NCR-9523958
 * 
 * DISCLAIMER:
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL
 * MULTIMEDIA COMMUNICATIONS LAB, BOSTON UNIVERSITY OR THE CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 * Contributors:
 *
 * Rajesh Krishnan
 */

#ifndef __SUBSCRIBER_HPP__
#define __SUBSCRIBER_HPP__

#include "list.hpp"
#include <limits.h>
#include <math.h>

// Stream statuses
enum Request {PLAY, PAUSE, FFWD, RWIND, QUIT, UNKWN};

class Virtualstream;

class Subscriber:public Node
{
  friend class Subscriberlist;
protected:
  int user_id;
  Request request;
  double duration;
  Virtualstream *mystream;
  
  // For queueing statistics purposes
public:
  static int num_subscriber;

  double start_wait_time;
  double end_wait_time;
  double wait_until;

  Subscriber(int id) 
    {user_id=id; request=PLAY; mystream = NULL; duration=DBL_MAX; \
     start_wait_time = end_wait_time = wait_until = -1.0; \
     num_subscriber++;};
  int GetUID(void) { return user_id;};
  void SetMystream(Virtualstream *s);
  Virtualstream *GetMystream(void) {return mystream;};
  Request GetRequest(void) {return request;};
  double GetDuration(void) {return duration;};
  void SetRequest(Request r) {request = r;};
  void SetDuration(double d) {duration = d;};
  void Passivate(void) {request = PLAY; duration = DBL_MAX;};
  void ResetWaitStats(void) { start_wait_time = end_wait_time = wait_until = -1.0;};
  virtual ~Subscriber();
  virtual void Dumpdata(void); 
  void SetWaitUntil();
};

class Subptr:public Node
{
public:
  static int num_subptr;
  Subscriber * sptr;
  Subptr(Subscriber *s) {sptr = s;num_subptr++;};
  ~Subptr();
  virtual void Dumpdata(void) {if(sptr) sptr->Dumpdata();};
};

class Subscriberlist:public List
{
public:
  Subscriberlist() {head = tail = 0; num_nodes = 0;};
  Subscriberlist(int n);
  ~Subscriberlist() {};
  Subscriber *ExtractRandom(void);
  Subscriber *RandomChoice(void);
};

#endif //__SUBSCRIBER_HPP__



