/* -*- Mode:C++ -*- */
/*
 * Copyright (c) Multimedia Communications Lab, Boston University
 *
 * Simulation Program to Evaluate Service Aggregation in VOD
 * =========================================================
 *
 * This work was supported by NSF under grant No. NCR-9523958
 * 
 * DISCLAIMER:
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL
 * MULTIMEDIA COMMUNICATIONS LAB, BOSTON UNIVERSITY OR THE CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 * Contributors:
 *
 * Wang Ke
 */

#include <iostream.h>
#include <math.h>
#include <string.h>
#include <assert.h>
#include "adins.hpp"
#include "config.hpp"

// whether to show some messages
extern int verbose;

// used to build the binary tree
int LAST_MERGE = 1;
int NOT_LAST_MERGE = 0;

// this was created to count the number of objects. For memory check purposes
extern void CheckObjects();

// simulation time k - see main.cpp
extern double k;

// ad insertion object - see main.cpp
extern Ad_Insertion global_ad_ins;


// translate the NStatus from integer to string
void print_NStts(NStatus nstts)
{
  if (verbose) {

    if (Config::debug_predict) {

      switch(nstts) {
      case (NORMAL):
	cout << " NORMAL ";
	break;
      case (NSADINS):
	cout << " NSADINS ";
	break;
      case (INVALID):
	cout << " INVALID ";
	break;
      case (ACCEL):
	cout << " ACCEL ";
	break;
      default:
	cout << " not translated ";
	break;
      }
    }
    else {
      switch(nstts) {
      case (NORMAL):
	cout << "nstts(" << (int)nstts <<") = NORMAL";
	break;
      case (NSADINS):
	cout << "nstts(" << (int)nstts << ") = NSADINS";
	break;
      case (ACCEL):
	cout << "nstts(" << (int)nstts << ") = ACCEL";
	break;
      case (INVALID):
	cout << "nstts(" << (int)nstts << ") = INVALID";
	break;
      default:
	cout << "nstts(" << (int)nstts << ") not translated";
	break;
      }
      cout << endl;
    }
  }
}

void print_Stts(Status sts) {

  int stts = (int)sts;

  switch(stts) {
    
  case (RTOPLAY):
    if (Config::debug_predict) 
      cout << " RTOPLAY ";
    else 
      cout << "stts(" << stts << ") = RTOPLAY";
    break;
  case (FASTPLAY):
    if (Config::debug_predict)
      cout << " FASTPLAY ";
    else
      cout << "stts(" << stts << ") = FASTPLAY";
    break;
  case (PLAYING):
    if (Config::debug_predict) 
      cout << " PLAYING ";
    else 
      cout << "stts(" << stts << ") = PLAYING";
    break;
  case (AD_INS):
    cout << "stts(" << stts << ") = AD_INS";
    break;
  case (ADINS_STT):
    cout << "stts(" << stts << ") = ADINS_STT";
    break;
  case (TRAILING_STT):
    cout << "stts(" << stts << ") = TRAILING_STT";
    break;
  case (PLAYING_STT):
    cout << "stts(" << stts << ") = PLAYING_STT";
    break;
  case (ADI_ADINS_COM):
    cout << "stts(" << stts << ") = ADI_ADINS_COM";
    break;
  case (MERGED_STT):
    cout << "stts(" << stts << ") = MERGED_STT";
    break;
  case (ADI_ADONLY_COM):
    cout << "stts(" << stts << ") = ADI_ADONLY_COM";
    break;
  case (ADI_PLAYING_COM):
    cout << "stts(" << stts << ") = ADI_PLAYING_COM";
    break;
  case (PL_PLAYING_COM):
    cout << "stts(" << stts << ") = PL_PLAYING_COM";
    break;
  case (ADI_VIDEO_AD_COM):
    cout << "stts(" << stts << ") = ADI_VIDEO_AD_COM";
    break;
  case (UNKNOWN):
    if (Config::debug_predict) 
      cout << " UNKNOWN ";
    else 
      cout << "stts(" << stts << ") = UNKNOWN";
    break;
  case (PAUSED):
    if (Config::debug_predict)
      cout << " PAUSED ";
    else
      cout << "stts(" << stts << ") = PAUSED";
    break;
  case (RWDING):
    if (Config::debug_predict)
      cout << " RWDING ";
    else 
      cout << "stts(" << stts << ") = RWDING";
    break;
  case (FWDING):
    if (Config::debug_predict) 
      cout << " FWDING ";
    else
      cout << "stts(" << stts << ") = FWDING";
    break;
  case (ATEOMV):
    if (Config::debug_predict)
      cout << " ATEOMV ";
    else 
      cout << "stts(" << stts << ") = ATEOMV";
    break;
  case (INBATCH):
    if (Config::debug_predict)
      cout << " INBATCH ";
    else 
      cout << "stts(" << stts << ") = INBATCH";
    break;
  default:
    cout << "stts(" << stts << ") not translated";
    break;
  }
  if (!Config::debug_predict)
    cout << endl;
}


// Print STREAM statistics
void print_STREAM(STREAM *sts) 
{
  if (verbose) {
    cout << "Stream id: " << sts->stream_id << "\tposition: " << sts->position
	 << "\nmovie_id: " << sts->movie_id << endl
	 << "ad_to_go: " << sts->ad_to_go << "\ttime_to_go: " 
	 << sts->time_to_go << "\nvideo_to_go: " << sts->video_to_go 
	 << "\tlength: " << sts->movie_length << endl;
  }
}

////////////////////////////////////////////////////////////
// This is the same as compar in util.c except for the 
// different type cast
//
// Returns the leading stream (largest position in the movie
//  stream) as the first element of an array given to qsort
////////////////////////////////////////////////////////////

int compare_streams(const void *stream1, const void *stream2)
{
  const Stream *s1 = (const Stream *)stream1;
  const Stream *s2 = (const Stream *)stream2;

  if (s1->movie_id != s2->movie_id)
    return (s1->movie_id - s2->movie_id);
  if (s2->position >= s1->position) 
    return 1;
  else
    return -1;
}



///////////////////////////////////////////////////////////
// compare cluster sizes. The same as comparclust
// Effect on qsort is that it returns the largest 
// cluster as the first element of the array
//////////////////////////////////////////////////////////

int compare_clusters(const void *cluster1, const void *cluster2)
{
  const Cluster *c1 = (const Cluster *)cluster1;
  const Cluster *c2 = (const Cluster *)cluster2;

  return ( (c2->last_member - c2->first_member) -
	   (c1->last_member - c1->first_member));
}



//////////////////////////////////////////////////////
// TriangMatrix class. Used to store the triangular 
// matrices needed in the optimal tree computation.
// Uses linear array of size n(n+1)/2 where n is the
// size of the matrix's side.
//////////////////////////////////////////////////////

template<class T>
TriangMatrix<T>::TriangMatrix()
{ 
  side = 0;
  max_side = 0;
  max_num_elem = 0;
  elem = NULL;
  num_tmatrix++;
}


// initialize matrix to a maximum size
template<class T>
void TriangMatrix<T>::init_matrix(int mxsd)
{
  if (elem != NULL) deallocate_matrix();
  if (mxsd > 0) {
    side = mxsd;
    assert ( (side *(side+1)/2) > 0);
    max_num_elem = (side*(side+1))/2;
    max_side = mxsd;
    init_vec(max_num_elem);
   
    if (elem == NULL) {
      cout << "ERROR: TriangMatrix::init - Memory exhausted\n";
      exit(1);
    }
    reset_matrix(mxsd);
  }
}

// initialize the array containing the matrix
template<class T>
void TriangMatrix<T>::init_vec(int mne)
{
  elem = new T[mne];
  assert(elem);
}

// deallocate memory
template<class T>
void TriangMatrix<T>::deallocate_matrix()
{
  if (elem != NULL) {
    delete [] elem;
    elem = NULL;
    max_num_elem = 0;
    max_side = 0;
    side = 0;
  }
}

// reassign matrix's "side" to newside
template<class T>
void TriangMatrix<T>::reset_matrix(int newside)
{
  if ( (elem != NULL) && (max_num_elem >= 0)) 
    memset(elem, 0, max_num_elem*sizeof(T));

  if ( (newside >= 0) && (newside <= max_side))
    side = newside;
  else{
#ifdef DEBUG
    cout<<"TriangMatrix<T>::reset_matrix to new side " << newside << "\n\n";
#endif
    side = 0; 
  }
}

//deallocate the matrix
template<class T>
TriangMatrix<T>::~TriangMatrix()
{
  deallocate_matrix();
  num_tmatrix--;
}

// get value stored in the matrix at position (i,j)
template<class T>
T TriangMatrix<T>::get_value(int i, int j)
{
  assert((j < side) && (i <= j) && (i >= 0) &&(j >= 0));

  if (elem == NULL) {
    cout << "ERROR: TriangMatrix::get_elem - elem array NULL\n";
    exit(1);
  }
  return elem[i*side+j-(i*(i+1)/2)];
}

// access the element (i,j) of the matrix.
template<class T>
T& TriangMatrix<T>::get_obj(int i, int j)
{
  assert((j < side) && (i <= j) && (i >= 0) &&(j >= 0));

  if (elem == NULL) {
    cout << "ERROR: " << "TriangMatrix::get_elem - elem array NULL\n";
    exit(1);
  }
  return elem[i*side+j-(i*(i+1)/2)];
}

// set value for element (i,j) of the matrix to "t"
template<class T>
void TriangMatrix<T>::set_value(int i, int j, T t)
{
  assert((j < side) && (i <= j) && (i >= 0) &&(j >= 0));

  if (elem == NULL) {
    cout << "ERROR: " <<"TriangMatrix::set_elem - elem array NULL\n";
    exit(1);
  }
  elem[i*side+j-(i*(i+1)/2)] = t;
}



////////////////////////////////////////////////////
// Stream object is used by ad_insertion object in
// computation procedures
////////////////////////////////////////////////////

Stream::Stream()
{
  position = stream_id = (double)UNINITIALIZED;
  movie_id = (int)UNINITIALIZED;
  movie_length = (double)UNINITIALIZED;

  merging_with = merging_point = cost = (double) UNINITIALIZED;
  ad_to_go = time_to_go = (double)UNINITIALIZED;
  video_to_go = (double) ZERO_VALUE;

  status = UNKNOWN;
  last_command_prepended = UNINITIALIZED;
  num_stream++;
}


Stream::~Stream()
{
  num_stream--;
}


// initialize Stream object from STREAM struct
int Stream::init(STREAM *sts, double window)
{
  return (this->init(sts->stream_id, sts->position, sts->movie_id, 
		     sts->movie_length, map_status(sts->status), sts->ad_to_go,
		     sts->time_to_go, sts->video_to_go, window));
}


int Stream::init(double str_id, double pos, int movie, double mv_len, 	     
		 Status stts, double atg, double ttg, double vtg,
		 double window)
{

  if ((str_id < 0.0) || (pos < 0.0) ||  (movie < 0) ||  (mv_len <= 0.0) ||
      (atg < 0.0) || (ttg < 0.0) || (window <= 0.0) ) {
    return ADI_FAILURE;
  }

  position = pos;
  stream_id = str_id;
  movie_id = movie;
  movie_length = mv_len;
  ad_to_go = atg;
  time_to_go = ttg;
  video_to_go = vtg;
  cost = max(mv_len - position,0.0);
  last_command_prepended = max_ncommands-1; 

  return ADI_SUCCESS;
}

// Store the different states the stream must go through, starting from the last one
void Stream::prepend_command(Status stts, double till_position)
{
  if (last_command_prepended < (max_ncommands-1)) {
    if (commands[last_command_prepended+1].status == stts)
      return;
  }

  if (last_command_prepended < 0) {
    cout << "ERROR: " <<"Stream::prepend_commands - stream "
	 <<stream_id << " command array overflow\n";
    exit(1);
  }

  commands[last_command_prepended].status = stts;
  commands[last_command_prepended].position_after_command = till_position;

  last_command_prepended--;
}

// map NStatus to Status
Status Stream::map_status(NStatus nstts)
{
  switch(nstts) {

  case (NORMAL):
    return (PLAYING);
    break;
  case (NSADINS):
    return (AD_INS);
    break;
  default:
    return (UNKNOWN);
  }
}

// for debugging purposes: print the commands calculated for the stream
void Stream::printCommands()
{
  for (int i = 0; i< max_ncommands;i++) {
    cout << "commands["<<i<<"].status = " << (int)commands[i].status << " - ";
    print_Stts(commands[i].status);
    cout << "for: "<< commands[i].duration <<endl;
    cout << "until position: "<<commands[i].position_after_command << endl;
  }
}

// cost without ad_insertion/clustering: movie length - initial position
double Stream::get_original_cost()
{
  return (max(0.0,(movie_length - position)));
}

// compute the cost for the stream with ad insertion 
double Stream::get_cost()
{
  double cost = 0.0;
  for (int i= 0; i < max_ncommands; i++ ) {
    cost = cost + cost_per_command(&commands[i]);
  }
  return cost;
}

double Stream::cost_per_command(Command *cmd)
{
  double cost = 0.0;

  switch(cmd->status) {
  case (MERGED_STT):
    cost = 0.0;
    break;
  default:
    cost = cmd->duration;
    break;
  }
  return cost;
}


// Print all stream's data. For debugging purposes
void Stream::printData(int i)
{
  cout << "\n ++++++++++++++++++++++++++++++++\nStream["<<i<<"]\n";
  printData();
  cout << "\n ++++++++++++++++++++++++++++++++\n";
}

void Stream::printData()
{
  int i;

  cout << "\nstream_id = " << stream_id <<"\nmovie = "<<movie_id << \
    "\nposition = " << position << endl;

  cout << "merging with " << merging_with << " at ("<< merging_point << \
       ") with cost:" << cost << endl;
  cout << "ad_to_go:" << ad_to_go << "  time_to_go:"<<time_to_go \
       << "  video_to_go:" << video_to_go << "  mv_len:" << movie_length
       << endl;

  if (last_command_prepended < (max_ncommands-1)) {
    for (i = last_command_prepended; i< max_ncommands; i++) {
      cout << "command["<<i<<"].status = " << (int)commands[i].status << " - ";
      print_Stts(commands[i].status);
      cout << "until position: "<<commands[i].position_after_command << endl;
    }
  }
}


///////////////////////////////////////////////////////////////
// Object used to store information regarding the different
// clusters
///////////////////////////////////////////////////////////////

Clusterer::Clusterer()
{
  num_clusterer++;
}

Clusterer::~Clusterer()
{
  num_clusterer--;
}

void Clusterer::sort_streams(Stream *streams, int nstreams)
{
  qsort(streams, nstreams, sizeof(Stream), compare_streams);
}


// Computes the point s1 and s2 converges. "clusteringOrExtreme" == 1 AND
// Config::check_eomv==1 imply that if the convergence point of s1 and s2
// is beyond the eomv, then converge_at returns failure
//
// s1 leads, s2 trails

PositionNode Clusterer::converge_at(Stream *s1, Stream *s2, 
				    double interval, int clusteringOrExtreme)
{
  PositionNode p1;

  // not watching the same movie. Can't converge
  if (s1->movie_id != s2->movie_id) {
    p1.position = (double)ADI_FAILURE;
    return p1;
  }

  // s1 has finished, should not be taken into consideration
  if ( s1->position >= s1->movie_length) {
    p1.position = (double)ADI_FAILURE;
    return p1;
  }

  // s2 has finished, should not be taken into consideration
  if ( s2->position >= s2->movie_length) {
    p1.position = (double)ADI_FAILURE;
    return p1;
  }

  // If s1 is a stream that has been in PLAYING state up
  // to now, then it might have video_to_go negative 
  // (indicating a non-initialized video state), in which case
  // it should be allocated the full amount of video time
  // needed
  if (s1->video_to_go < 0) {
    assert((s1->ad_to_go == AMAX));
    assert((s1->time_to_go == TPRIME));
    s1->video_to_go = MAX_VMIN;
  }

  // QoS problem. Left ad_to_go is greater than
  // the overall time_to_go left for the stream.  
  if ((s1->time_to_go < s1->ad_to_go)) {
    cout << "s1->time_to_go smaller than s1->ad_to_go" << endl;
    s1->printData();
    s1->printCommands();
    s2->printData();
    s2->printCommands();
    assert((s1->time_to_go >= s1->ad_to_go)); // force interruption
  }

  double difference = s1->position - s2->position;
  int ntimes=0;
  double rest=0.0;
  double tmpdiff=0.0;

  if (difference == 0.0) {

    // Resulting stream should be restricted by the highest
    // QoS constraints of s1 and s2.
    Stream *ts = (s1->time_to_go <= s2->time_to_go) ? s1 : s2;
    p1.video_to_go = ts->video_to_go;
    p1.time_to_go = ts->time_to_go;
    p1.ad_to_go = ts->ad_to_go;
    p1.position = ts->position;

    // sanity check. Time_to_go _has_ to be greater.
    if (p1.time_to_go < p1.ad_to_go) {
      cout << "p1.time_to_go = " << p1.time_to_go << endl
	   << "p1.ad_to_go = " << p1.ad_to_go << endl
	   << "ntimes = " << ntimes << endl
	   << "rest = " << rest << endl
	   << "difference = " << difference << endl;
      ts->printData();
      ts->printCommands();
      assert((p1.time_to_go >= p1.ad_to_go)); // force interruption
    }
    return p1;
  }

  if (difference <= s1->time_to_go) { 
    
    if (difference <= s1->ad_to_go) {
      p1.position = s1->position;
      p1.ad_to_go = s1->ad_to_go - difference;
      p1.time_to_go = s1->time_to_go - difference;
      p1.video_to_go = s1->video_to_go;
    }
    else {
      
      // Error: remaining video time (video_to_go) smaller than
      // the last video segment time. 
      if ((s1->video_to_go < LAST_VIDEO_TIME)) {
	cout << "ERROR: s1 video to go inconsistent\ns1: leading:\n";
	s1->printData();
	cout << "s2: trailing:\n";
	s2->printData();
	assert((s1->video_to_go >= LAST_VIDEO_TIME)); // force interruption
      }
      
      // the variable ntimes below indicates how many VMIN segments
      // s1 must go through. tmpvtg tracks the remainder of video_to_go
      // after ntimes segments of VMIN.
      double tmpvtg = modulus((s1->video_to_go - LAST_VIDEO_TIME),VMIN);
      tmpvtg = (tmpvtg == 0.0) ? VMIN:tmpvtg;

      tmpdiff = difference;
      difference = difference - s1->ad_to_go;
      
      ntimes = (int)ceil(difference/AMAX);
      rest = modulus(difference,AMAX);

      // sanity check. ntimes == 0 iff difference == 0.
      if ((ntimes == 0)) {
	cout << "ERROR: ntimes == 0\n";
	s1->printData();
	cout << "s2: trailing:\n";
	s2->printData();
	assert(ntimes); // force interruption
      }

      p1.position = s1->position + tmpvtg + ((ntimes-1)*VMIN);

      p1.video_to_go = s1->video_to_go - (((ntimes-1)*VMIN) + tmpvtg);
      p1.time_to_go = s1->time_to_go - (difference + s1->ad_to_go);
      p1.ad_to_go = AMAX - rest;

      // If there was enough to catch up after an ad time, then 
      // changes are needed
      if (rest == 0) {
	p1.ad_to_go = 0;
      }
    }

    if (p1.time_to_go < p1.ad_to_go) {
      cout << "tmpdiff = " << tmpdiff << endl
	   << "p1.position = " << p1.position << endl
	   << "p1.time_to_go = " << p1.time_to_go << endl
	   << "p1.ad_to_go = " << p1.ad_to_go << endl
	   << "ntimes = " << ntimes << endl
	   << "rest = " << rest << endl
	   << "difference = " << difference << endl;
      s1->printData();
      s1->printCommands();
      assert((p1.time_to_go >= p1.ad_to_go)); // force interruption
    }
  }
  else {

    difference = difference - s1->time_to_go;
    ntimes = (int)ceil(difference/TPRIME);
    rest = modulus(difference,TPRIME);

    rest = ((rest == 0) ? TPRIME:rest);

    int ntimes_2 = (int)ceil(rest/AMAX);
    double rest_2 = modulus(rest,AMAX);
    
    p1.position = s1->position + s1->video_to_go + (ntimes-1)*(MAXT-TPRIME) +
      (ntimes_2 -1)*VMIN;
    
    if (rest == TPRIME) { // when difference is a multiple of TPRIME
      p1.time_to_go = 0;
      p1.ad_to_go = 0;
      p1.video_to_go = LAST_VIDEO_TIME;
    }
    else {
      p1.time_to_go = TPRIME - rest;
      p1.video_to_go = (MAX_VMIN - ((ntimes_2 - 1)*VMIN));
      if (rest_2 == 0) {
	p1.ad_to_go = 0;
      }
      else {
	p1.ad_to_go = AMAX - rest_2;
      }
    }
    
    if (p1.time_to_go < p1.ad_to_go) {
      cout << "p1.time_to_go = " << p1.time_to_go << endl
	   << "p1.ad_to_go = " << p1.ad_to_go << endl
	   << "ntimes = " << ntimes << endl
	   << "rest = " << rest << endl
	   << "ntimes_2 = " << ntimes_2 << endl
	   << "rest_2 = " << rest_2 << endl
	   << "difference = " << difference << endl;
      s1->printData();
      s1->printCommands();
      assert((p1.time_to_go >= p1.ad_to_go)); // force interruption
    }
  }

  if ( (p1.position - s2->position) > interval) {  
    p1.position = (double) ADI_FAILURE;
  }

  if ( (Config::check_eomv) && (clusteringOrExtreme)) {
    if ( (p1.position >= s2->movie_length)) {
      p1.position = (double) ADI_FAILURE;
    }
  }
  return p1;
}

void Clusterer::cluster(Stream *streams, int nstreams, 
		       Cluster *clusters,
		       double window, 
		       int *nchannels, int *nclusters)
{
  // Clusterer::cluster - virtual function - needs to be derived
  ;
}


//////////////////////////////////////////////////////////////////
// Clustering based on EMCL
//

EMCL_Clusterer::EMCL_Clusterer()
  :Clusterer()
{
  num_emcl++;
}

EMCL_Clusterer::~EMCL_Clusterer()
{
  num_emcl--;
}

void EMCL_Clusterer::cluster(Stream *streams, int nstreams, 
			    Cluster *clusters, double window, 
			    int *nclusters, int *nchannels)
{
  int i, j;
  PositionNode p1;

  sort_streams(streams, nstreams);

  for (i=1, j=0, *nclusters = 0; i<nstreams; i++) {

    p1 = converge_at(&streams[j], &streams[i],window, 1);
    if ( (streams[i].movie_id == streams[j].movie_id) &&
	 p1.position>(double)ADI_FAILURE) {
      continue;
    }
    else {

    }

    // old condition: if (j < (i-1)) - need to change to j <= (i-1) so that 
    // "clusters" of 1 stream can be included. The reason is that later on
    // the build command set is called on each cluster.

    if (j <= (i-1)) {
      clusters[*nclusters].first_member = j;
      clusters[*nclusters].last_member = i-1;
      (*nclusters)++;
    }
    j = i;
  }

  if (j<= (i-1)) { // getting even single streams here

    clusters[*nclusters].first_member = j;
    clusters[*nclusters].last_member = i-1;
    (*nclusters)++;
  }

  if((*nclusters) == 0) {
    cout << "WARNING: no clusters formed in EMCL_Clusterer!!\n\n";
    return;
  }

  qsort(clusters, *nclusters, sizeof(Cluster), compare_clusters);

  for (i=0, *nchannels = 0; i < *nclusters; i++) 
    *nchannels += clusters[i].last_member - clusters[i].first_member;
}
  
  
/////////////////////////////////////////////////////
// Ad_Insertion object. Implements the ad insertion
// algorithm
////////////////////////////////////////////////////

Ad_Insertion::Ad_Insertion(Clusterer *clust)
{
  assert(clust != NULL);
  clusterer = clust;
  streams = NULL;
  clusters = NULL;
  
  nstreams = nclusters = nchannels = max_cluster_element = 0;
  TIMEBASE = (double) UNINITIALIZED;

}


Ad_Insertion::~Ad_Insertion()
{
  deallocate();
}

// Based on the number of STREAM structs, allocates the number
// of streams and clusters
void Ad_Insertion::allocate(STREAM *sts, int num_streams, double window)
{
  assert(num_streams > 0);

  deallocate();

  nstreams = num_streams;

  if ( (streams = new Stream[num_streams]) == NULL) {
    cout << "ERROR: " 
	 <<"Ad_Insertion::allocate - memory exhausted for streams\n";
    exit(1);
  }

  // worst case scenario - each stream in its own cluster
  nclusters = num_streams;
  
  if ( nclusters > 0 ) {

    if ( (clusters = new Cluster[nclusters]) == NULL) {
      cout << "ERROR: " <<"Ad_Insertion::allocate -" 
	   << " memory exhausted for clusters\n";
      exit(1);
    }
  }
      
  for (int i = 0; i< nstreams; i++) {

    if (streams[i].init(&sts[i], window) == ADI_FAILURE) {
      cout << "ERROR: stream " << sts[i].stream_id << " failed init step\n";
      cout << "Stream stats: "; print_STREAM(&sts[i]);
      exit(1);
    }
  }
}


void Ad_Insertion::deallocate()
{
  if (streams != NULL) {
    delete [] streams;
    streams = NULL;
  }

  if (clusters != NULL) {
    delete [] clusters;
    clusters = NULL;
  }
}

// find original cost of the set of streams without ad insertion mechanisms
double Ad_Insertion::get_original_cost()
{
  double ocost = 0.0, ccost = 0.0;

  for (int i=0;i<nclusters;i++) {
    ccost = 0.0;
    for (int j=clusters[i].first_member; j<=clusters[i].last_member;j++){
      ccost += streams[j].get_original_cost();
    }
    ocost +=ccost;
  }
  return ocost;
}

// cost of the set of streams with ad insertion. Computed based
// on the cost of the last stream of the cluster
double Ad_Insertion::get_cost()
{
  double ncost = 0.0;

  for (int i=0; i < nclusters; i++) {
    ncost += streams[clusters[i].last_member].cost;
  }
  return ncost;
}

// Cost of the streams. Computed based on the set of commands they must go through
double Ad_Insertion::cost_for_commands()
{
  double cost = 0.0, ccost = 0.0;
  int i,j;
  for (i=0; i < nclusters; i++) {
    for (j = clusters[i].first_member; j <= clusters[i].last_member;j++){
      ccost += streams[j].get_cost();
    }
    cost +=ccost;
    ccost = 0.0;
  }
  return cost;
}
    
// invoke the ad insertion policy on the streams
void Ad_Insertion::invoke_policy(STREAM *sts, int num_streams, 
				 double window, double current_time)
{
  int i,tree_elem, cur_cluster;

  if (num_streams <=0) {
    return;
  }

  allocate(sts, num_streams, window);

  clusterer->cluster(this->streams, this->nstreams, this->clusters, window,
		     &(this->nclusters), &(this->nchannels));

  // this defines the largest number of elements any cluster has
  max_cluster_element = clusters[0].last_member - clusters[0].first_member+1;
  tree_elem  = ((max_cluster_element)*(max_cluster_element+1))/2;

  assert(tree_elem>0);

  // These 3 trees should be used solely for this 
  t_tree.init_matrix(max_cluster_element);
  position_tree.init_matrix(max_cluster_element);
  cost_tree.init_matrix(max_cluster_element);

  for (i=0;i<nclusters;i++) {

    cur_cluster = clusters[i].last_member - clusters[i].first_member +1;
    tree_elem = ((cur_cluster)*(cur_cluster+1))/2;
    assert(tree_elem > 0);

    // This is actually redundant for the first time, but we need to 
    // differentiate between init and reset.

    cost_tree.reset_matrix(cur_cluster);
    position_tree.reset_matrix(cur_cluster);
    t_tree.reset_matrix(cur_cluster);

    set_one_cluster(i,window);
  }
  return;
}

// Compute everything inside one cluster
void Ad_Insertion::set_one_cluster(int which, double window)
{
  PositionNode *posn = NULL;
  Stream *sts = NULL;
  Stream *trails = &streams[clusters[which].last_member];
  
  int number_nodes = clusters[which].last_member - \
    clusters[which].first_member+1;
  int first_member = clusters[which].first_member;

  int i,j, m;

  // initialization of leaf nodes
  for (i=0, j=first_member;i<number_nodes;i++, j++) {

    if (streams[j].stream_id == Config::stream_id)
      j+=0;
    
    sts = &streams[j];
    posn = &position_tree.get_obj(i,i);
    posn->position = sts->position;
    posn->ad_to_go = sts->ad_to_go;
    posn->time_to_go = sts->time_to_go;
    posn->video_to_go = sts->video_to_go;
    
    t_tree.set_value(i,i,i);
    cost_tree.set_value(i,i, (sts->cost));
  }
  
  // to compute the 3 trees for the algorithm
  for (m = 1; m < (number_nodes); m++) { // skipping the initialization step
    for (i = 0; i< (number_nodes-m); i++) {
      if ( (i==0) && (m==(number_nodes-1)) ) 
	compute_merge_point(i, i+m, first_member, window,1);
      else 
	compute_merge_point(i, i+m, first_member, window, 0);
      set_min_cost_tree(i,i+m, first_member, trails->movie_length);
    }
  }

  // everything is built. Start building stream commands

  trails->prepend_command(PLAYING_STT, min((trails->position+window),
					   trails->movie_length));
  trails->cost = cost_tree.get_value(0, number_nodes-1);

  build_paths(0, number_nodes-1 ,first_member, LAST_MERGE);

  for (i= 0, j = first_member; \
	 j <= clusters[which].last_member; i++, j++) {
    if ( (streams[j].stream_id == Config::stream_id) 
	 || (streams[j].stream_id == Config::snd_stream)) {
      i+=0;
    }
    build_stream_cmd_path(i, first_member);
  }

  // to give the trailing stream the correct value for the last 
  // ad_to_go and time_to_go and video_to_go, we must obtain the last
  // stream that merged with the trailing stream, this is done by finding
  // the midpoint of the first and last stream
  i = t_tree.get_value(0, number_nodes-1);

  trails->merging_point = streams[i+first_member].merging_point;
  trails->ad_to_go = streams[i+first_member].ad_to_go;
  trails->time_to_go = streams[i+first_member].time_to_go;
  trails->video_to_go = streams[i+first_member].video_to_go;
}

// Checking elements in TriangMatrix. For debugging purposes
void Ad_Insertion::checkTriangMatrix()
{
  cerr << "AD:CT: cost_tree = " << cost_tree.max_num_elem
       << "\nAD:CT: t_tree = " << t_tree.max_num_elem
       << "\nAD:CT: position_tree = " << position_tree.max_num_elem << endl;
}


// Assigns states to streams lead and trail. Inevitably the leading stream
// has to be assigned ADINS_STT and the trailing stream PLAYING_STT
void Ad_Insertion::assign2(Stream *lead, Stream *trail, PositionNode *p, 
			   int last_merge)
{
  lead->merging_with = trail->stream_id;
  lead->merging_point = p->position;
  lead->prepend_command(ADINS_STT,p->position);
  trail->prepend_command(PLAYING_STT, p->position);

  if (last_merge) {
    trail->ad_to_go = p->ad_to_go;
    trail->time_to_go = p->time_to_go;
    trail->video_to_go = p->video_to_go;
    if (trail->stream_id == Config::stream_id) {
      cout << "Assign2 for stream " << trail->stream_id 
           << " at time " << TIMEBASE << "\nLeading stream:"
           << lead->stream_id << endl;
      lead->printData();
      cout << "\nassigning a:" << p->ad_to_go << " t:" 
           << p->time_to_go << " v:" << p->video_to_go
           << " merging:" << p->position;
    }      
  }
}

// Debugging the trees built 
void Ad_Insertion::debug_tree_struct()
{
  for (int i=0;i<nclusters;i++) {
    cout << "\nDebugging cluster " << i << endl;
    debug_tree_struct(i);
  }
}
    
void Ad_Insertion::debug_tree_struct(int which_cluster)
{
  debug_tree_struct(0,clusters[which_cluster].last_member -
		    clusters[which_cluster].first_member, 
		    clusters[which_cluster].first_member);
}

void Ad_Insertion::debug_tree_struct(int leading, int trailing, int first)
{
  assert(leading <= trailing);
  PositionNode *p1 = &position_tree.get_obj(leading,trailing);
  int midpoint;

  cout << "\n\nStreams " << streams[leading+first].stream_id << " and " 
       << streams[trailing+first].stream_id 
       << "\nmerging at: " << p1->position 
       << "\n\tad_to_go: " << p1->ad_to_go 
       << "\n\ttime_to_go: " << p1->time_to_go 
       << "\n\tvideo_to_go: " << p1->video_to_go
       << "\ncost: " << cost_tree.get_value(leading, trailing) << endl;
	streams[leading+first].printData();
   cout << "Stream " << streams[leading+first].stream_id
        << " command cost:"<< streams[leading+first].get_cost()<<endl;
        streams[trailing+first].printData();
   cout << "Stream " << streams[trailing+first].stream_id
        << " command cost:"<< streams[trailing+first].get_cost()<<endl;
  cout << "Next...\n\n";
  if (leading < trailing) {
    midpoint = t_tree.get_value(leading, trailing);
    cout << "middle stream id: " << streams[midpoint+first].stream_id << endl;
  
    debug_tree_struct(leading, midpoint, first);
    debug_tree_struct(midpoint+1, trailing, first);
  }

  cout << "leaf node. End\n";
}

// Based on the binary tree (which streams merge with which) assign the
// states the streams will have to go through    
void Ad_Insertion::build_paths(int leading, 
			       int trailing, 
			       int first_stream, 
			       int last_merge)
{
  assert(leading <= trailing);

  if (leading == trailing) // reached leaf node
    return;

  int midpoint = t_tree.get_value(leading, trailing);

  assign2(&streams[midpoint+first_stream], &streams[trailing+first_stream],
	  &position_tree.get_obj(leading, trailing), last_merge);

  streams[midpoint+first_stream].cost = cost_tree.get_value(leading, midpoint);

  build_paths(leading, midpoint, first_stream, LAST_MERGE);
  build_paths(midpoint+1, trailing, first_stream, NOT_LAST_MERGE);

}

// Obtain the merging point of 2 streams
void Ad_Insertion::compute_merge_point(int leading, int trailing, 
				       int first_stream, double window,
				       int extreme)
{
  assert (leading < trailing);

  PositionNode merging_point = clusterer->converge_at( \
    &streams[leading+first_stream], &streams[trailing+first_stream],window,extreme);

  PositionNode *p1;

  if (merging_point.position <= (double)ADI_FAILURE) {
    cout<<"Ad_Insertion::compute_merge_point - ADI_FAILURE for " 
	<< leading+first_stream << " and " << trailing+first_stream 
	<< "\nPrinting data:\n";
    streams[leading+first_stream].printData();
    streams[trailing+first_stream].printData();
    cout << "ERROR: " <<"Ad_Insertion::compute_merge_point " 
	 <<"ADI_FAILURE unable to proceed";
    assert(0);
  }

  p1 = &position_tree.get_obj(leading, trailing);
  p1->position = merging_point.position;
  p1->ad_to_go = merging_point.ad_to_go;
  p1->time_to_go = merging_point.time_to_go;
  p1->video_to_go = merging_point.video_to_go;

}

// Amount of ad that stream lead(offset in a cluster)+f_str(first stream
// index in a cluster) needs in order to reach position length
double Ad_Insertion::compute_amount_ad_needed(int lead,
					    int f_str, double length)
{
  Stream *leading = &streams[lead+f_str];
  double amount_needed = 0.0;
  if (length < leading->position) {
    cerr << "SET_MIN_COST: Wrong assumption - stream " << leading->stream_id
	 << " movie length " << length 
         << " less than position of stream " << leading->position;
  }
  else if (length == leading->position) {
    amount_needed = 0.0;
  }
  else if (length < (leading->position + leading->video_to_go)) {

    double pointX = leading->position - ((MAXT-TPRIME) - leading->video_to_go);
    double video_amount = length - pointX;
        
    int namaxfromx =((int)ceil(video_amount/VMIN)) -1;
    int namaxtop = max(((int)ceil((leading->position-pointX)/VMIN)-1),0);
    amount_needed = leading->ad_to_go + (namaxfromx - namaxtop)*AMAX;
  }
  else if (length == (leading->position + leading->video_to_go)) {
    amount_needed = leading->time_to_go;
  }
  else {

    int  first_h = (int)ceil((length - 
			      (leading->position + leading->video_to_go))/ 
			     (MAXT-TPRIME));
    double first_p = (first_h-1) * TPRIME;
    
    double second_p = length - (leading->position+leading->video_to_go+
				((first_h-1)*(MAXT-TPRIME)));
    int second_h = (int)ceil(second_p/VMIN);
    
    if (second_h > Config::ratio_tprime_amax) {
      second_h = min(second_h,Config::ratio_tprime_amax);
    }
    second_h = max((second_h-1),0);
    amount_needed = leading->time_to_go + first_p + (second_h*AMAX);
  }
  return amount_needed;
}      
    
// Obtain the tree that has minimum cost
void Ad_Insertion::set_min_cost_tree(int leading, int trailing, 
				     int f_str, double length)
{
  double min=0.0, temp=0.0, ccost=0.0, amount_o=0.0;
  double tmp_amount=0.0, amount =0.0;
  int i,opt;

  assert( leading < trailing);

  ccost = max((length - position_tree.get_obj(leading, trailing).position),
	      0.0);

  // If we are considering here clustering beyond end of movie, then the
  // cost computed should NOT include parts of the movie that are beyond
  // the end of movie. If the movie is merging with another one, the
  // cost should be computed from the point leading merges with leading+1
  // This procedure repeated for all elements will ensure the cost minimization
  // in path construction

  if (position_tree.get_obj(leading,trailing).position > length) {
    amount = compute_amount_ad_needed(leading,f_str,length);
    amount_o = 0.0;

    tmp_amount = \
      (position_tree.get_obj(leading, leading+1).position>length)? 0:
      (streams[leading+f_str].position - streams[leading+1+f_str].position);

    tmp_amount = amount-tmp_amount;
  }
  else {
    amount_o = (position_tree.get_obj(leading,leading).position -        
		position_tree.get_obj(trailing, trailing).position);
    tmp_amount = 0.0;
  }

  min = cost_tree.get_value(leading, leading) +
    cost_tree.get_value(leading+1, trailing) + amount_o +
    tmp_amount - ccost;    

  opt = leading;					    
  for (i= leading+1; i<trailing; i++) {

    if (position_tree.get_obj(i,trailing).position > length) {
      tmp_amount = \
	(position_tree.get_obj(leading, i).position>length)? 0:
	(streams[leading+f_str].position - streams[i+f_str].position);
      tmp_amount = amount - tmp_amount;
      amount_o = 0.0;
    }
    else {
      amount_o = (position_tree.get_obj(i,i).position -        
		  position_tree.get_obj(trailing, trailing).position);
      tmp_amount = 0.0;
    }
    
    temp = cost_tree.get_value(leading, i) + 
      cost_tree.get_value(i+1,trailing) + amount_o +
      tmp_amount -
      ccost;

    if (temp < min) {
      min = temp;
      opt = i;
    }
  }
  
  cost_tree.set_value(leading, trailing, min);
  t_tree.set_value(leading, trailing, opt);
}

// Returns the NStatus of stream with id stream_id at time at_time
NStatus Ad_Insertion::get_stream_status(double stream_id, 
					double at_position,
					double at_time) 
{
  int i = find_stream(stream_id);
  double time_offset = at_time - TIMEBASE;
  double time_past = 0.0;
  double predicted_position = 0.0;
  int j = -1;

  if ( i == NOT_FOUND) {
    return INVALID;
  }

  if ( time_offset < 0.0) {
    return INVALID;
  }

  while(time_past <= time_offset) {
    j++;
    if ( j >= Stream::max_ncommands) {
      return INVALID;
    }

    if (streams[i].commands[j].status == MERGED_STT) {
      return ( get_stream_status(streams[i].merging_with, 
				 at_position,  
				 at_time));
    }
    
    if (streams[i].commands[j].status == TRAILING_STT) {
      predicted_position = get_position_at_time(i,j,(time_offset-time_past));

      if ( (predicted_position >= streams[i].movie_length) &&
	   (fabs(predicted_position - at_position) <= AMIN)) {
	return (NSATEOMV);
      }

      if (fabs(predicted_position - at_position) > AMIN) {
	return(INVALID);
      }
      return (NORMAL);
    }

    time_past += streams[i].commands[j].duration;
  }

  time_past -= streams[i].commands[j].duration;

  if (fabs(get_position_at_time(i,j, time_offset - time_past) 
	   - at_position) > AMIN) {
    return(INVALID);
  }

  return( get_status_from_command(streams[i].commands[j].status, 
				  (time_offset-time_past)));
}

// auxiliary command used to obtain the status from within a command
NStatus Ad_Insertion::get_status_from_command(Status stts, double offset)
{
  NStatus s;
  switch (stts) {
  case (ADI_ADONLY_COM):
    s = NSADINS;
    break;
  case (ADI_PLAYING_COM):
    s = NORMAL; // PLAYING;
    break;
  case (PL_PLAYING_COM):
    s = NORMAL; //PLAYING;
    break;
  case (ADI_VIDEO_AD_COM):
    if (modulus(offset, (AMAX+VMIN)) < VMIN)
      s = NORMAL; //PLAYING;
    else 
      s = NSADINS; //AD_INS;
    break;
  case (ADI_ADINS_COM):    
    if ( (modulus(offset,MAXT)) >= ((R_TPRIME_AMAX *(VMIN + AMAX)) - VMIN)) {
      s = NORMAL; //PLAYING;
    }
    else {
      if ( modulus(modulus(offset, MAXT), (AMAX+VMIN)) >= AMAX ) {
	s = NORMAL; //PLAYING;
      }
      else {
	s = NSADINS; //AD_INS;
      }
    }
    break;
  case (TRAILING_STT):
    s = NORMAL; //PLAYING;
    break;
  default:
    s = INVALID; //UNKNOWN;
    break;
  }
  return s; 
}

// Here offset is the amount of time that is spent inside 
// streams[str_index].commands[cmd_index]

double Ad_Insertion::get_position_at_time(int str_index,
					  int cmd_index, 
					  double offset)
{
  double dd, temp;
  Status stts = streams[str_index].commands[cmd_index].status;

  switch (stts) {
  case (ADI_ADONLY_COM):
    dd = 0.0;
    break;
  case (ADI_PLAYING_COM):
    dd = offset;
    break;
  case (PL_PLAYING_COM):
    dd = offset;
    break;
  case (ADI_VIDEO_AD_COM):
    dd = modulus(offset,(AMAX+VMIN));
    if (dd < VMIN)
      dd = dd + (floor(offset/(AMAX+VMIN))*VMIN);
    else 
      dd = floor(offset/(AMAX+VMIN))*VMIN + VMIN;
    break;
  case (ADI_ADINS_COM):    
    dd = modulus(offset, MAXT);
    temp = (R_TPRIME_AMAX * (VMIN+AMAX))-VMIN;

    if ( dd >= temp) {
      dd = (dd-temp) + ((R_TPRIME_AMAX -1) * VMIN) + \
	(floor(offset/MAXT)*(MAX_VMIN));
    }
    else {
      dd = (floor(dd/(AMAX+VMIN))*VMIN) + (floor(offset/MAXT)*MAX_VMIN);
      temp =  modulus(modulus(offset, MAXT),(AMAX+VMIN)); 
      if (temp >= AMAX) {
	dd = (temp-AMAX) + dd;
      }
    }
    break;
  case (TRAILING_STT):
    dd = offset; //PLAYING;

    // Here it is possible to "cut short" because we see an offset
    // into the TRAILING_STT state. TRAILING_STT is the state of the
    // trailing stream of a cluster. Probably it is being tried a 
    // probe of the stream's status after the stream reached EOMV
    if (offset >= streams[str_index].commands[cmd_index].duration)
      dd = streams[str_index].commands[cmd_index].duration;
    break;
  default:
    return (-2.0);
    break;
  }

  if (cmd_index == 0) {
    temp = streams[str_index].position;
  }
  else {
    temp = streams[str_index].commands[cmd_index-1].position_after_command;
  }
  
  return (dd+temp); 
}


// sts is the stream index within the cluster, f_stream is the 
// global index of the first stream within the same cluster

void Ad_Insertion::build_stream_cmd_path(int sts, int f_stream)
{
  Stream *str = &streams[sts+f_stream];
  if (str->stream_id == Config::stream_id ) {
    str->stream_id +=0.0;
  }
  int i,j;
  int n_command_built = 0;
  int n_long, n_short;
  int nsegments = ((str->last_command_prepended+1) == \
		   (str->max_ncommands-1)) ? 1:2;
  double points[MAX_ARRAY_STTS];
  double rest, restlong = 0.0, restamax = 0.0;
  double inipos = 0.0;
  Status ss[MAX_ARRAY_STTS];
  PositionNode *pn1;

  assert ( (points != NULL) && (ss != NULL));
  assert ( (str->last_command_prepended + nsegments) == \
	   (str->max_ncommands-1));

  points[0] = str->position;
  ss[0] = str->status; // right now considered irrelevant. Maybe used later

  for (i=0; i<nsegments; i++) {
    points[i+1] = \
      str->commands[i+str->last_command_prepended+1].position_after_command;
    ss[i+1] = str->commands[i+str->last_command_prepended+1].status;
  }

  for (i=1; i< (nsegments+1); i++) {

    switch(ss[i]) {
      // There are only 2 possible states at most. One is PLAYING_STT and the
      // other is ADINS_STT. Possible state sequences are PLAYING_STT, 
      // PLAYING_STT+ADINS_STT, and ADINS_STT. All ADINS_STT must begin with 
      // ADONLY_COM, even if this means an ad_only time of 0 (in case the 
      // ad_to_go time is 0). And after ADINS_STT is reached, that is the 
      // final stream state, after which the only possible command is the 
      // MERGED.

    case (PLAYING_STT):
      str->commands[n_command_built].status = PL_PLAYING_COM;
      str->commands[n_command_built].position_after_command = points[i];
      str->commands[n_command_built].duration = (points[i]-points[i-1]);
      if (str->commands[n_command_built].duration < 0) {
	cout << "ERROR: duration negative " 
	     << str->commands[n_command_built].duration << endl;
	  str->printData();
	  str->printCommands();
	assert(str->commands[n_command_built].duration >= 0);
      }
      n_command_built++;
      break;
      

    case (ADINS_STT): // occurs only after PLAYING_STT state
      j = find_stream(str->merging_with);
      assert(j >= (sts+f_stream)); 
      // j should always be bigger than sts for j is trailing

      pn1 = &position_tree.get_obj(sts,j-f_stream);
      rest = str->position - streams[j].position; 
      assert(rest>=0);

      if (rest == 0) {
	assert(i==1);
      }
      if ( (str->time_to_go == 0.0) ) {
	assert(str->ad_to_go == 0.0);
	double tmpvtg1 = (str->video_to_go < 0.0) ? 0.0:str->video_to_go;
	str->commands[n_command_built].status= ADI_PLAYING_COM;
	str->commands[n_command_built].duration = tmpvtg1;
	inipos = str->commands[n_command_built].position_after_command = \
	  (points[i-1] + tmpvtg1);
	n_command_built++;
	str->ad_to_go = AMAX;
	str->time_to_go = TPRIME;
      }
      else if ( (str->time_to_go > 0.0) && (str->ad_to_go == 0.0)) {
	if ( (str->video_to_go >= 0.0) && (str->video_to_go <= LAST_VIDEO_TIME)) {
	  cerr << " Assertion wrong: str " << str->stream_id
	       << " with t:" << str->time_to_go
	       << " a:0.0 and v:" << str->video_to_go << endl;
	  
	  cout << "Assertion wrong: str " << str->stream_id
	       << " with t:" << str->time_to_go
	       << " a:0.0 and v:" << str->video_to_go << endl;
	  str->printData();
	  assert(0);
	}

	if (str->video_to_go < 0.0) {
	  cerr << " Assertion wrong: str " << str->stream_id
	       << " with video_to_go < 0.0 and ad_to_go == 0.0\n";
	  cout << " Assertion wrong: str " << str->stream_id
	       << " with video_to_go < 0.0 and ad_to_go == 0.0\n";
	  str->printData();
	  assert(0);
	}

	double tmpvtg2 = modulus( (str->video_to_go-LAST_VIDEO_TIME),VMIN);
	tmpvtg2 = ((tmpvtg2 == 0.0) ? VMIN:tmpvtg2);

	str->video_to_go = str->video_to_go - tmpvtg2;

	str->commands[n_command_built].status= ADI_PLAYING_COM;
	str->commands[n_command_built].duration = tmpvtg2;
	inipos = str->commands[n_command_built].position_after_command = \
	  (points[i-1] + tmpvtg2);
	n_command_built++;
	str->ad_to_go = min(str->time_to_go, AMAX);
      }
      else {
	inipos = points[i-1];
      }

      str->commands[n_command_built].status = ADI_ADONLY_COM;
      str->commands[n_command_built].position_after_command = inipos;

      if (fabs(rest) <= str->ad_to_go) {
	str->commands[n_command_built].duration = fabs(rest);
	if (str->commands[n_command_built].duration < 0) {
	  cout << "ERROR: duration negative " 
	       << str->commands[n_command_built].duration << endl;
	  str->printData();
	  str->printCommands();
	  assert(str->commands[n_command_built].duration >= 0);
	}
	str->commands[n_command_built+1].status = MERGED_STT;
	str->commands[n_command_built+1].duration = 0;
	str->commands[n_command_built+1].position_after_command = \
	  str->commands[n_command_built].position_after_command;
	n_command_built +=2;
	return;
      }
      
      str->commands[n_command_built].duration = str->ad_to_go;
      if (str->commands[n_command_built].duration < 0) {
	cout << "ERROR: duration negative " 
	     << str->commands[n_command_built].duration << endl;
	str->printData();
	str->printCommands();
	assert(str->commands[n_command_built].duration >= 0);
      }
      n_command_built++;

      if ( fabs(rest) <= str->time_to_go) {
	n_short = (int) floor((rest- str->ad_to_go)/AMAX);

	if (n_short == 0 ) {

	  str->commands[n_command_built].status = ADI_PLAYING_COM;
	  str->commands[n_command_built].position_after_command =      \
	    str->commands[n_command_built-1].position_after_command +  \
	    VMIN; //tmpvtg;
	  str->commands[n_command_built].duration = VMIN; //tmpvtg;
	  n_command_built++;

	  str->commands[n_command_built].status = ADI_ADONLY_COM;
	  str->commands[n_command_built].duration =                    \
	    modulus((rest - str->ad_to_go), AMAX);
	  if (str->commands[n_command_built].duration < 0) {
	    cout << "ERROR: duration negative " 
		 << str->commands[n_command_built].duration << endl;
	    str->printData();
	    str->printCommands();
	    assert(str->commands[n_command_built].duration >= 0);
	  }
	  str->commands[n_command_built].position_after_command =      \
	    str->commands[n_command_built-1].position_after_command;
	  n_command_built++;
	
	}
	else {
	  str->commands[n_command_built].status = ADI_VIDEO_AD_COM;
	  str->commands[n_command_built].duration = n_short*(VMIN+AMAX);
	  if (str->commands[n_command_built].duration < 0) {
	    cout << "ERROR: duration negative " 
		 << str->commands[n_command_built].duration << endl;
	    str->printData();
	    str->printCommands();
	    assert(str->commands[n_command_built].duration >= 0);
	  }
	  str->commands[n_command_built].position_after_command = \
	    str->commands[n_command_built-1].position_after_command +  \
	    n_short*(VMIN);
	  n_command_built++;

	  restlong = modulus((rest - str->ad_to_go), AMAX);

	  if (restlong != 0 ) {
	    // it means that (rest-str->ad_to_go) is not divisible by AMAX
	    // Thus the merging point is actually somewhere in between an
	    // ADONLY command 

	    str->commands[n_command_built].status = ADI_PLAYING_COM;
	    str->commands[n_command_built].duration = VMIN;
	    str->commands[n_command_built].position_after_command =     \
	      str->commands[n_command_built-1].position_after_command + \
	      VMIN;
	    n_command_built++;

	    str->commands[n_command_built].status = ADI_ADONLY_COM;
	    str->commands[n_command_built].duration = restlong;
	    if (str->commands[n_command_built].duration < 0) {
	      cout << "ERROR: duration negative " 
		   << str->commands[n_command_built].duration << endl;
	      str->printData();
	      str->printCommands();
	      assert(str->commands[n_command_built].duration >= 0);
	    }

	    str->commands[n_command_built].position_after_command =     \
	      str->commands[n_command_built-1].position_after_command;
	    n_command_built++;
	  }
	}

	str->commands[n_command_built].status = MERGED_STT;
	str->commands[n_command_built].duration = 0;
	str->commands[n_command_built].position_after_command =        \
	  str->commands[n_command_built-1].position_after_command;
	n_command_built++;
	return;
      }

      n_short = (int) floor((str->time_to_go - str->ad_to_go)/AMAX);
      str->commands[n_command_built].status = ADI_VIDEO_AD_COM;
      str->commands[n_command_built].position_after_command =           \
	str->commands[n_command_built-1].position_after_command +       \
	n_short*VMIN;
      str->commands[n_command_built].duration = (VMIN+AMAX)* n_short;
      if (str->commands[n_command_built].duration < 0) {
	cout << "ERROR: duration negative " 
	     << str->commands[n_command_built].duration << endl;
	str->printData();
	str->printCommands();
	assert(str->commands[n_command_built].duration >= 0);
      }

      str->commands[n_command_built+1].status = ADI_PLAYING_COM;
      str->commands[n_command_built+1].duration = LAST_VIDEO_TIME;

      str->commands[n_command_built+1].position_after_command =         \
	str->commands[n_command_built].position_after_command +         \
	str->commands[n_command_built+1].duration;
      n_command_built +=2;


      rest -= str->time_to_go;

      restlong = modulus(rest,TPRIME);
      n_long = (int) floor(rest/TPRIME);

      // For the rest, if restlong == 0.0, it means that the merging point
      // occurs at the end of a full TPRIME cycle, in which case we need
      // to consider a sequence of ADI_VIDEO_AD_COM instead of ADI_ADINS_COM
      // One way of solving the problem is to subtract one full TPRIME cycle
      // and assign TPRIME to restlong.

      if (restlong == 0.0) {
	restlong = TPRIME;
	n_long--;
      }


      if (n_long > 0) {
	str->commands[n_command_built].status = ADI_ADINS_COM;
	str->commands[n_command_built].position_after_command =         \
	  str->commands[n_command_built-1].position_after_command +     \
	  (n_long *MAX_VMIN);
	str->commands[n_command_built].duration = n_long * MAXT;
	n_command_built++;
      }
      
      n_short = (int) floor(restlong/AMAX);

      str->commands[n_command_built].status = ADI_ADONLY_COM;

      if (n_short == 0) {
	str->commands[n_command_built].duration = restlong;
	if (str->commands[n_command_built].duration < 0) {
	  cout << "ERROR: duration negative " 
	       << str->commands[n_command_built].duration << endl;
	  str->printData();
	  str->printCommands();
	  assert(str->commands[n_command_built].duration >= 0);
	}
	str->commands[n_command_built].position_after_command =         \
	  str->commands[n_command_built-1].position_after_command;

	str->commands[n_command_built+1].status = MERGED_STT;
	str->commands[n_command_built+1].duration = 0.0;
	str->commands[n_command_built+1].position_after_command =       \
	  str->commands[n_command_built].position_after_command;
	return;
      }

      str->commands[n_command_built].duration = AMAX;
      str->commands[n_command_built].position_after_command =           \
	str->commands[n_command_built-1].position_after_command;
      n_command_built++;

      n_short--;

      str->commands[n_command_built].status = ADI_VIDEO_AD_COM;
      str->commands[n_command_built].duration = n_short * (VMIN+AMAX);
      if (str->commands[n_command_built].duration < 0) {
	cout << "ERROR: duration negative " 
	     << str->commands[n_command_built].duration << endl;
	str->printData();
	str->printCommands();
	assert(str->commands[n_command_built].duration >= 0);
      }
      str->commands[n_command_built].position_after_command =           \
	str->commands[n_command_built-1].position_after_command +       \
	(n_short * VMIN);
      n_command_built++;

      restamax = modulus(restlong, AMAX);

      if (restamax == 0) {
	str->commands[n_command_built].status = MERGED_STT;
	str->commands[n_command_built].duration = 0.0;
	str->commands[n_command_built].position_after_command =         \
	  str->commands[n_command_built-1].position_after_command;
	return;
      }

      str->commands[n_command_built].status = ADI_PLAYING_COM;
      str->commands[n_command_built].duration = VMIN;
      str->commands[n_command_built].position_after_command =           \
	str->commands[n_command_built-1].position_after_command + VMIN;
      n_command_built++;

      str->commands[n_command_built].status = ADI_ADONLY_COM;
      str->commands[n_command_built].duration = restamax;
      if (str->commands[n_command_built].duration < 0) {
	cout << "ERROR: duration negative " 
	     << str->commands[n_command_built].duration << endl;
	str->printData();
	str->printCommands();
	assert(str->commands[n_command_built].duration >= 0);
      }
      str->commands[n_command_built].position_after_command =           \
	str->commands[n_command_built-1].position_after_command;

      str->commands[n_command_built+1].status = MERGED_STT;
      str->commands[n_command_built+1].duration = 0.0;
      str->commands[n_command_built+1].position_after_command =         \
	str->commands[n_command_built].position_after_command;

      return;
      break;
    default:
      cout << "ERROR: Ad_Insertion:build_stream_cmd_path: wrong stt[" \
	   << i << "] = " << (int)ss[i] << endl;
      exit(1);
      break;
    }
  }
  // it can only reach here if it has not had an ADINS_STT, and therefore
  // must be the trailing stream of the cluster

  if (str->commands[n_command_built-1].status == PL_PLAYING_COM) {
    str->commands[n_command_built].status = TRAILING_STT;
    str->commands[n_command_built].position_after_command = \
      str->commands[n_command_built-1].position_after_command;
    str->commands[n_command_built].duration = 0;
  }
  return;
}


int Ad_Insertion::find_stream(double stream_id)
{
  int i = 0, j = 0;
  for (i=0;i<nstreams;i++) {
    if (streams[i].stream_id == stream_id){
      j = 1;
      break;
    }
  }
  
  if (j == 0) {
    return NOT_FOUND;
  }

  return i;
}


void Ad_Insertion::print_stream_cmd()
{
  for (int i = 0; i < nstreams; i++) {
    cout << "\nStream[" << i <<"]:\n";
    streams[i].printData();
    streams[i].printCommands();
  }
}

void Ad_Insertion::debug_streams()
{
    debug_streams(0, nstreams-1);
}

void Ad_Insertion::debug_streams(int kk, int l)
{
  int i;
  for (i = kk; i <= l; i++) {
    streams[i].printData(i);
    streams[i].printCommands();
  }
}

void Ad_Insertion::debug_stream_id(double id)
{
  int i;
  if ((i = find_stream(id)) != NOT_FOUND) {
    debug_streams(i,i);
  }
  else {
    cout << "AD_INS: stream " << id << " not found\n";
  }
}

void Ad_Insertion::debug_chain_id(double start)
{
  int  i;

  i = find_stream(start);
  if (i != NOT_FOUND) {
    debug_streams(i,i);
    while (streams[i].merging_with > -1) {
      debug_stream_id(streams[i].merging_with);
      i = find_stream(streams[i].merging_with);
      if (i == NOT_FOUND) {

	cout << "ERROR: stream (" << streams[i].merging_with
	     << ") not found. Debug interrupted\n";
	break;
      }
    }
  }

  cout << "ERROR: stream (" << start
       << ") not found. Debug interrupted\n";
}
void Ad_Insertion::debug_clusters()
{
  int nstrs = 0;

  for (int i = 0; i < nclusters; i++ ) {

    cout << "\ncluster[" << i <<"]\n";
    clusters[i].printData();
    nstrs = nstrs + (clusters[i].last_member - clusters[i].first_member) +1;
  }
  cout << "\nTotal number clustered: " << nstrs << endl;
  cout << endl;

}

double Ad_Insertion::debug_merge_with(double streamId)
{
  double last_one = -1.0;
  for (int i=0; i< nstreams; i++) {
    if (streams[i].merging_with == streamId) {
      cout << "Found stream " << streams[i].stream_id << " merging with "
	   << streamId << endl;
      streams[i].printData();
      streams[i].printCommands();
      last_one = streams[i].stream_id;
      cout << endl;
    }
  }
  return last_one;
}

double Ad_Insertion::to_merge_with(double stid)
{
  int i = find_stream(stid);
  if (i == NOT_FOUND)
    return STREAM_NOT_FOUND;
  else 
    return streams[i].merging_with;
}
