/* -*- Mode:C++ -*- */
/*
 * Copyright (c) Multimedia Communications Lab, Boston University
 *
 * Simulation Program to Evaluate Service Aggregation in VOD
 * =========================================================
 *
 * This work was supported by NSF under grant No. NCR-9523958
 * 
 * DISCLAIMER:
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL
 * MULTIMEDIA COMMUNICATIONS LAB, BOSTON UNIVERSITY OR THE CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 * Contributors:
 *
 * Prithwish Basu
 */

#include <math.h>
#include <stdio.h>
#include <limits.h>
#include <stdlib.h>
#include "stream.h"


/***************************************************************************/
/* Clustering algorithm -- implements EMCL(x)                              */
/***************************************************************************/
void cluster(STREAM *streams, int nstreams, CLUSTER *clusters, double trec, 
	    int *nclusters, int *nchannels)
{
    int i, j;

    if (nstreams == 1) 
    {
	*nchannels = 1;
	*nclusters = 0;
	return;
    }

    /* Sort streams -- can use counting sort, but I am lazier than thou */
    qsort(streams, nstreams, sizeof(STREAM), compar);

    /* Find all clusters and their count */
    for(i=1, j=0, *nclusters=0; i<nstreams; i++)
    {
	if(CARRY_SAME_PROGRAM(streams[i].movie_id, streams[j].movie_id)
	   && (Kfactor*(streams[j].position - streams[i].position) <= trec)) 
	    continue;
			
	if(j < i-1)
	{
	    clusters[*nclusters].firstmemind = j;
	    clusters[*nclusters].lastmemind = i-1;
	    (*nclusters)++;
	}
	j = i;
    }

    if(j < i-1)
    {
	clusters[*nclusters].firstmemind = j;
	clusters[*nclusters].lastmemind = i-1;
	(*nclusters)++;
    }

    if(!(*nclusters)) return;

    /* Sort clusters */
    qsort(clusters, *nclusters, sizeof(CLUSTER), comparclust);


    for(i=0, *nchannels=0; i<*nclusters; i++) 
	*nchannels += clusters[i].lastmemind - clusters[i].firstmemind; 

}
