/* -*- Mode:C++ -*- */
/*
 * Copyright (c) Multimedia Communications Lab, Boston University
 *
 * Simulation Program to Evaluate Service Aggregation in VOD
 * =========================================================
 *
 * This work was supported by NSF under grant No. NCR-9523958
 * 
 * DISCLAIMER:
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL
 * MULTIMEDIA COMMUNICATIONS LAB, BOSTON UNIVERSITY OR THE CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 * Contributors:
 *
 * Prithwish Basu
 */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "rsma_heur.h"
#include "rsma.h"


extern double RSMA_cost;

PQUEUE *deletept(PQUEUE *elt, POINT node)
{
    int flag=0;
    PQUEUE *pt, *del;

    for (pt = elt; pt!=NULL; pt=pt->next)
    {
	if ((pt->node.x == node.x) && (pt->node.y == node.y))
	{
	    flag = 1;
	    if ((pt->prev==NULL)&&(pt->next==NULL)) 
		elt = NULL;		
	    else if (pt->prev==NULL)
	    {
		pt->next->prev = NULL;
		elt = pt->next;
	    }
	    else if (pt->next==NULL)
		pt->prev->next = NULL;
	    else 
	    {
		pt->prev->next = pt->next;
		pt->next->prev = pt->prev;
	    }
	    del = pt;
	}
    }
    if (flag) free(del);
	
    return elt;
} 

PQUEUE *add(PQUEUE *elt, PQUEUE *pnode)
{
    PQUEUE *pt;
	    
    for (pt = elt; pt!=NULL; pt=pt->next)
    {
	if (pt->node.x >= pnode->node.x)
	{
	    if (pt->prev==NULL)
	    {
		pt->prev = pnode; pnode->next = pt; pnode->prev = NULL; elt = pnode;
	    }
	    else if (pt->next==NULL)
	    {
		pnode->next = pt; pnode->prev = pt->prev; pt->prev->next = pnode; pt->prev = pnode;
	    }
	    else 
	    {
		pt->prev->next = pnode; pnode->prev = pt->prev; pt->prev = pnode; 
		pnode->next = pt;
	    }
	    break;
	}
	if (pt->next==NULL)
	{
	    pnode->next = NULL; pnode->prev = pt; pt->next = pnode; break;
	}
    }
    if (!elt)
    {
	elt = pnode;
	elt->prev = elt->next = NULL;
    }

    return elt;
}

void print_queue(PQUEUE *pq)
{
    PQUEUE *p;
    
    for (p=pq; p!=NULL; p=p->next)
	cout << p->node.x << "_" << p->node.y << "(id's=" << p->id[0] << "," << p->id[1] << ") ";

    cout << endl;;

}


double norm(POINT node)
{
    return (node.x + node.y);
}


/* This is a heuristic algorithm for finding the minimum RSA */
/* for any number/type of points on the grid, not necessarily on D_n */
/* In our problem, graph[len] is an array of points on D_n */ 
void compute_rsa_heur(POINT *graph, int len, int **out_tree)
{
    int i, *tree;
    double z=0, old=-1, this_cost;
    PQUEUE *p, *q, *Nset = NULL, *old_p, *old_q, *par, *gr;
    POINT pnode;

    /* this will denote the RSA tree */
    tree = (int *) calloc((len*(len+1))/2, sizeof(int));
    
    for (i=0; i<len; i++) 
    {	
	gr = (PQUEUE *) malloc(sizeof(PQUEUE));
	gr->node = graph[i];
	gr->next = gr->prev = NULL;
	gr->id[0] = len-i-1;
	gr->id[1] = len-i-1;
	Nset = add(Nset, gr);
	Nset->id[0] = len-i-1;
	Nset->id[1] = len-i-1;	
    }

    RSMA_cost = 0;
    
    while (Nset->next != NULL) 
    {
	old = -1;
	old_p = old_q = NULL;
	
	/* find a pair of nodes with maximum norm of their parent */
	for (p=Nset; p!=NULL; p=p->next, i++)
	{
	    for(q=p->next; q!=NULL; q=q->next)
	    {
		z = norm(parent(p->node, q->node));
		if (z>old) 
		{
		    old = z; old_p = p; old_q = q;
		}
	    }
	}

	pnode = parent(old_p->node, old_q->node);

	par = (PQUEUE *) malloc (sizeof(PQUEUE));
	par->node = pnode;
	par->id[0] = maximum(old_p->id[0], old_q->id[0]);
	par->id[1] = minimum(old_p->id[1], old_q->id[1]);
	par->prev = par->next = NULL;

	this_cost = old_p->node.y - pnode.y + old_q->node.x - pnode.x;
	RSMA_cost += this_cost;

 	int_assign(par->id[1], par->id[0], len, tree, 1);
	
	Nset = deletept(Nset, old_p->node);
	Nset = deletept(Nset, old_q->node);
	Nset = add(Nset, par); 
    }

    *out_tree = tree;
    
}
