/* -*- Mode:C++ -*- */
/*
 * Copyright (c) Multimedia Communications Lab, Boston University
 *
 * Simulation Program to Evaluate Service Aggregation in VOD
 * =========================================================
 *
 * This work was supported by NSF under grant No. NCR-9523958
 * 
 * DISCLAIMER:
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL
 * MULTIMEDIA COMMUNICATIONS LAB, BOSTON UNIVERSITY OR THE CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 * Contributors:
 *
 * Prithwish Basu
 */

#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include "stream.h"

extern void print_NStts(NStatus);

/***************************************************************************/
/* Integer comparison function for quicksort                               */
/***************************************************************************/
int compar(const void *va, const void *vb)
{
    const STREAM *a = (const STREAM *)va;
    const STREAM *b = (const STREAM *)vb;
    if(a->movie_id != b->movie_id) return (a->movie_id - b->movie_id);
    if (b->position >= a->position)
	return 1;
    else return -1;
}

/***************************************************************************/
/* Comparison function for quicksorting clusters                           */
/***************************************************************************/
int comparclust(const void *va, const void *vb)
{
    const CLUSTER *a = (const CLUSTER *)va;
    const CLUSTER *b = (const CLUSTER *)vb;
    return ((b->lastmemind - b->firstmemind) -
	    (a->lastmemind - a->firstmemind));
}

/***************************************************************************/
/* This is a simple populator function -- replace with own funky stuff     */ 
/* REPLACED WITH ZIPFIAN POPULARITY DISTRIBUTION                           */
/***************************************************************************/
#define _COMMENT___
#ifndef _COMMENT___
void populate(STREAM *streams, int nstreams, double plength, int nmovies)
{
    int i, j;
    double *movie_pops; /* Movie popularities */
    double rnd;

    /* Movie popularities follow Zipfian distribution */
    movie_pops = (double *) malloc(sizeof(double)*nmovies);
    for (i=1, movie_pops[0]=0; i<=nmovies; i++)
	movie_pops[0] += (double)1.0/(double)i;
    movie_pops[0] = 1.0/movie_pops[0];
    for (i=1; i<nmovies; i++)
	movie_pops[i] =  movie_pops[i-1] +  (movie_pops[0]/(double)(i+1));

    /* The program positions in every movie are uniformly distributed */
    for (i=0; i<nstreams; i++)
	{
	    rnd = unif_rand();
	    for (j=0; j<nmovies; j++)
		if (rnd<movie_pops[j]) break;
	    streams[i].movie_id = j;
	    streams[i].stream_id = i;
	    streams[i].position = (int) floor(plength * unif_rand()); 
	    streams[i].status = NORMAL;
	}
}
#endif


void output_streams(STREAM *stream, int len)
{
    int i;

    cout << "-----------------------------------------------------------" << endl;
    cout << "Stream Logs" << endl;
    for (i=0; i<len; i++) {
	cout << "stream_id = "<< stream[i].stream_id <<"\tMovie_id = " << stream[i].movie_id
	     << "\tPosition = " << stream[i].position << "\tStatus = ";
	print_NStts(stream[i].status); cout << endl; 
    }
    cout << "-----------------------------------------------------------" << endl;
}


/*************************************************************************/
/* Prints the stream id's in of streams in a particular cluster */
void print_stream_id(STREAM *s, CLUSTER c)
{
    int i;

    cout << "The streams in this cluster have IDs:" << endl;
    for (i=c.firstmemind; i<=c.lastmemind; i++)
/* 	cout << s[i].stream_id << " (" << s[i].position << ") "; */
	cout << s[i].stream_id << " ";
    cout << endl;
}


/*************************************************************************/
/* Prints the stream info */
void print_stream_info(STREAM s)
{
    cout << "The stream information:" << endl;
    cout << "Stream ID: " << s.stream_id 
	 << "\tPosition: " << s.position
	 << "\tMovie ID: " << s.movie_id
	 << "\tStatus: " << (int)s.status;
    cout << endl;
}


/*************************************************************************/
/* Find the array index from the stream id */
int find_stream_index(double stream_id, STREAM *stream, int nstreams)
{
    int i;
    
    for (i=0; i<nstreams; i++)
	if (stream[i].stream_id == stream_id)
	    return i;
    
    return -1;
}
/*************************************************************************/




