/* -*- Mode:C++ -*- */
/*
 * Copyright (c) Multimedia Communications Lab, Boston University
 *
 * Simulation Program to Evaluate Service Aggregation in VOD
 * =========================================================
 *
 * This work was supported by NSF under grant No. NCR-9523958
 * 
 * DISCLAIMER:
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL
 * MULTIMEDIA COMMUNICATIONS LAB, BOSTON UNIVERSITY OR THE CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 * Contributors:
 *
 * Rajesh Krishnan, 
 * Prithwish Basu,
 * Wang Ke
 */

#include "vod.hpp"
#include "config.hpp"

void Config::Dumpdata(void)
{
  cout << "VERBOSE = " << data.VERBOSE << "\n"; 
  cout << "MAXSIMCLOCK = " << data.MAXSIMCLOCK << "\n"; 
  cout << "SIMGRAIN = " << data.SIMGRAIN << "\n";
  cout << "ZIPF_PARAM = " << data.ZIPF_PARAM << "\n";
  cout << "MOVIE_DB = " << data.MOVIE_DB << "\n";
  cout << "TOT_USERS = " << data.TOT_USERS << "\n";
  cout << "MV_RQ_RATE = " << data.MV_RQ_RATE << "\n";
  cout << "QT_RQ_RATE = " << data.QT_RQ_RATE << "\n";
  cout << "FAST_PLAY_RATE = " << data.FAST_PLAY_RATE << "\n";
  cout << "FF_RATE = " << data.FF_RATE << "\n";
  cout << "RW_RATE = " << data.RW_RATE << "\n";
  cout << "BATCH_INTERVAL = " << data.BATCH_INTERVAL << "\n";
  cout << "RNG_RQ_RATE = " << data.RNG_RQ_RATE << "\n";
  cout << "FF_RQ_RATE = " << data.FF_RQ_RATE << "\n";
  cout << "FF_DUR = " << data.FF_DUR << "\n";
  cout << "RW_RQ_RATE = " << data.RW_RQ_RATE << "\n";
  cout << "RW_DUR = " << data.RW_DUR << "\n";
  cout << "PAU_RQ_RATE = " << data.PAU_RQ_RATE << "\n";
  cout << "PAU_DUR = " << data.PAU_DUR << "\n";
  cout << "CLUST_RECOMP = " << data.CLUST_RECOMP << "\n";
  
  cout << "STT_RANDOM = " << stt_random << endl;
  cout << "ALLOW_RECOMP = " << allow_recomp << endl;
  cout << "STT_STREAMS = " << stt_streams << endl;
  cout << "CLUST_WINDOW = " << clust_window << endl;
  cout << "MIN_AD_TIME = " << min_ad_time << endl;
  cout << "MAX_AD_TIME = " << max_ad_time << endl;
  cout << "MAX_AD/MIN_AD = " << ratio_max_min_ad << endl;
  cout << "MIN_VIDEO_TIME = " << min_video_time << endl;
  cout << "MIN_VIDEO/MIN_AD = " << ratio_vmin_amin << endl;
  cout << "LONG_TERM_AD_TIME = " << long_term_ad_time << endl;
  cout << "LONG_TERM_AD/MAX_AD = " << ratio_tprime_amax << endl;
  cout << "LONG_TERM_WINDOW = " << long_term_window << endl;
  cout << "MAX_VMIN = " << long_term_video << endl;
  cout << "INTERACTIONS = " << allowing_interactions << endl;
  cout << "AGGREGATE = " << aggregate << endl;
  cout << "RESET_TIME = " << reset_time << endl;
  cout << "PLOT_TIME_POS = " << plot_time_pos << endl;
  cout << "CHECK_EOMV = " << check_eomv << endl;
  cout << "STREAM_ID = " << stream_id << endl;
  cout << "SND_STREAM = " << snd_stream << endl;
  cout << "DEBUG_AT_RECOMP = " << debug_at_recomp << endl;
  cout << "JOIN_INFO = " << join_info << endl;
  cout << "USER_ID = " << user_id << endl;
  cout << "WAIT_DISTR_POWER = " << wait_distr_power << endl;
  cout << "WAIT_MAX_TIME = " << wait_max_time << endl;
  cout << "MAX_NUM_CHAN = " << max_num_chan << endl;
  cout << "PERC_OD = " << perc_OD << endl;
  cout << "RETURN_ONDEMAND = " << return_ondemand << endl;
  cout << "RECOMP_AT_EVENT = " << recomp_at_event << endl;
  cout << "COST_FORMULATION = " << (int)cost_formulation << endl;
  cout << "COST_RATE = " << cost_rate << endl;
  cout << "COST_MVLEN_MULTIPLIER = " << cost_mvlen_multiplier << endl;
  cout << "SHOW_USERS_PER_STREAM = " << show_users_per_stream << endl;
  cout << "SEED_A0 = " << seed_a0 << endl;
  cout << "SEED_A1 = " << seed_a1 << endl;
  cout << "SEED_A2 = " << seed_a2 << endl;
  cout << "PREDICT = " << predict << endl;
  cout << "DEBUG_PREDICT = " << debug_predict << endl;
  cout << "DEBUG_PREDICT_VALUE = " << debug_predict_value << endl;
  cout << "MAXHISTORY = " << MAXHISTORY << endl;
  cout << "SEPARATE_IN_CLUSTERS = " << data.SEPARATE_IN_CLUSTERS << endl;  
}

int read_config(char *fname, CFGDATA *d)
{
  FILE *fpconfig;
  char oneline[81];
  char *p, *v;
  int tmp_int;
  
  if((fpconfig = fopen(fname, "r")) == NULL) return(FAILURE);
  
  while(fgets(oneline, 80, fpconfig) != NULL)
	{
		p = strip_lwsc(oneline);
      
		// Skip Comment line
		if(*p == '#') continue;
      
		// If equal sign is missing skip line
		if((v = strchr(p, '=')) == NULL) continue;
      
		// Split Parameter and Value
		*v++ = '\0';
		p = strip_lwsc(p);
		v = strip_lwsc(v);
      
		if(!strcmp(p, "VERBOSE"))
		{d->VERBOSE = atoi(v); continue;}
      
		if(!strcmp(p, "MAXSIMCLOCK"))
		{d->MAXSIMCLOCK = atoi(v); continue;}
      
		if(!strcmp(p, "SIMGRAIN"))
		{sscanf(v, "%lf", &d->SIMGRAIN); continue;}
      
		if(!strcmp(p, "ZIPF_PARAM"))
		{sscanf(v, "%lf", &d->ZIPF_PARAM); continue;}
      
		if(!strcmp(p, "MOVIE_DB"))
		{strcpy(d->MOVIE_DB, v); continue;}
      
		if(!strcmp(p, "FAST_PLAY_RATE"))
		{sscanf(v, "%lf", &d->FAST_PLAY_RATE); continue;}
      
		if(!strcmp(p, "FF_RQ_RATE"))
		{sscanf(v, "%lf", &d->FF_RQ_RATE); continue;}
      
		if(!strcmp(p, "FF_DUR"))
		{sscanf(v, "%lf", &d->FF_DUR); continue;}
      
		if(!strcmp(p, "RW_RQ_RATE"))
		{sscanf(v, "%lf", &d->RW_RQ_RATE); continue;}
      
		if(!strcmp(p, "RW_DUR"))
		{sscanf(v, "%lf", &d->RW_DUR); continue;}
      
		if(!strcmp(p, "PAU_RQ_RATE"))
		{sscanf(v, "%lf", &d->PAU_RQ_RATE); continue;}
      
		if(!strcmp(p, "PAU_DUR"))
		{sscanf(v, "%lf", &d->PAU_DUR); continue;}
      
		if(!strcmp(p, "CLUST_RECOMP"))
		{sscanf(v, "%lf", &d->CLUST_RECOMP); continue;}
      
		if(!strcmp(p, "RNG_RQ_RATE"))
		{sscanf(v, "%lf", &d->RNG_RQ_RATE); continue;}
      
		if(!strcmp(p, "FF_RATE"))
		{d->FF_RATE = atoi(v); continue;}
      
		if(!strcmp(p, "RW_RATE"))
		{d->RW_RATE = atoi(v); continue;}
      
		if(!strcmp(p, "TOT_USERS"))
		{d->TOT_USERS = atoi(v); continue;}
      
		if(!strcmp(p, "MV_RQ_RATE"))
		{sscanf(v, "%lf", &d->MV_RQ_RATE); continue;}
      
		if(!strcmp(p, "QT_RQ_RATE"))
		{sscanf(v, "%lf", &d->QT_RQ_RATE); continue;}
      
		if(!strcmp(p, "BATCH_INTERVAL"))
		{sscanf(v, "%lf", &d->BATCH_INTERVAL); continue;}
      
		if(!strcmp(p, "AGG_POLICY"))
		{d->AGG_POLICY = atoi(v); continue;}
      
      
		if(!strcmp(p, "CLUST_WINDOW"))
		{sscanf(v, "%lf", &(Config::clust_window)); continue;}
      
		if(!strcmp(p, "USER_ID"))
		{sscanf(v, "%lf", &(Config::user_id)); continue;}
      
		if(!strcmp(p, "DEBUG_AT_RECOMP"))
		{sscanf(v, "%lf", &(Config::debug_at_recomp)); continue;}
      
		if(!strcmp(p, "RESET_TIME"))
		{sscanf(v, "%lf", &(Config::reset_time)); continue;}
      
		if(!strcmp(p, "AGGREGATE"))
		{sscanf(v, "%d", &(Config::aggregate)); continue;}
      
		if(!strcmp(p, "JOIN_INFO"))
		{sscanf(v, "%d", &(Config::join_info)); continue;}
      
		if(!strcmp(p, "STREAM_ID"))
		{sscanf(v, "%lf", &(Config::stream_id)); continue;}
      
		if(!strcmp(p, "SND_STREAM"))
		{sscanf(v, "%lf", &(Config::snd_stream)); continue;}
      
		if(!strcmp(p, "PLOT_TIME_POS"))
		{sscanf(v, "%d", &(Config::plot_time_pos)); continue;}
      
		if(!strcmp(p, "CHECK_EOMV"))
		{sscanf(v, "%d", &(Config::check_eomv)); continue;}
      
		if(!strcmp(p, "STT_RANDOM"))
		{sscanf(v, "%d", &(Config::stt_random)); continue;}
      
		if(!strcmp(p, "STT_DEMAND"))
		{sscanf(v, "%d", &(Config::stt_demand)); continue;}
      
		if(!strcmp(p, "ALLOW_RECOMP"))
		{sscanf(v, "%d", &(Config::allow_recomp)); continue;}
      
		if(!strcmp(p, "STT_STREAMS"))
		{sscanf(v, "%d", &(Config::stt_streams)); continue;}
      
		if(!strcmp(p, "MIN_AD_TIME"))
		{sscanf(v, "%lf", &(Config::min_ad_time)); continue;}
      
		if(!strcmp(p, "MAX_AD_TIME"))
		{sscanf(v, "%lf", &(Config::max_ad_time)); continue;}
      
		if(!strcmp(p, "MAX_AD/MIN_AD"))
		{sscanf(v, "%d", &(Config::ratio_max_min_ad)); continue;}
      
		if(!strcmp(p, "INTERACTIONS"))
		{sscanf(v, "%d", &(Config::allowing_interactions)); continue;}
      
		if(!strcmp(p, "MIN_VIDEO_TIME"))
		{sscanf(v, "%lf", &(Config::min_video_time)); continue;}
      
		if(!strcmp(p, "MIN_VIDEO/MIN_AD"))
		{sscanf(v, "%d", &(Config::ratio_vmin_amin)); continue;}
      
		if(!strcmp(p, "LONG_TERM_AD_TIME"))
		{sscanf(v, "%lf", &(Config::long_term_ad_time)); continue;}
      
		if(!strcmp(p, "LONG_TERM_AD/MAX_AD"))
		{sscanf(v, "%d", &(Config::ratio_tprime_amax)); continue;}
      
		if(!strcmp(p, "LONG_TERM_WINDOW"))
		{sscanf(v, "%lf", &(Config::long_term_window)); continue;}
      
		if(!strcmp(p, "LONG_TERM_VIDEO"))
		{sscanf(v, "%lf", &(Config::long_term_video));continue;}
      
		if(!strcmp(p, "WAIT_DISTR_POWER"))
		{sscanf(v, "%lf", &(Config::wait_distr_power));continue;}
      
		if(!strcmp(p, "WAIT_MAX_TIME"))
		{sscanf(v, "%lf", &(Config::wait_max_time));continue;}
      
		if(!strcmp(p, "PERC_OD"))
		{sscanf(v, "%lf", &(Config::perc_OD));continue;}
      
		if(!strcmp(p, "MAX_NUM_CHAN"))
		{sscanf(v, "%d", &(Config::max_num_chan));continue;}

		if(!strcmp(p, "RETURN_ONDEMAND"))
		{sscanf(v, "%d", &(Config::return_ondemand));continue;}

		if(!strcmp(p, "RECOMP_AT_EVENT"))
		{sscanf(v, "%d", &(Config::recomp_at_event));continue;}

		if(!strcmp(p, "COST_FORMULATION"))
		{
			sscanf(v, "%d", &(tmp_int));

			switch(tmp_int) {
			case(0): Config::cost_formulation = BIT; break;
			case(1): Config::cost_formulation = TIME_L_INF; break;
			case(2): Config::cost_formulation = TIME_L_FIN; break;
			};

			continue;
		}

		if(!strcmp(p, "COST_RATE"))
		{sscanf(v, "%lf", &(Config::cost_rate));continue;}

		if(!strcmp(p, "COST_MVLEN_MULTIPLIER"))
		{sscanf(v, "%d", &(Config::cost_mvlen_multiplier));continue;}

		if(!strcmp(p, "SHOW_USERS_PER_STREAM"))
		{sscanf(v, "%d", &(Config::show_users_per_stream));continue;}

		if(!strcmp(p, "SEED_A0"))
		{sscanf(v, "%hu", &(Config::seed_a0));continue;}

		if(!strcmp(p, "SEED_A1"))
		{sscanf(v, "%hu", &(Config::seed_a1));continue;}

		if(!strcmp(p, "SEED_A2"))
		{sscanf(v, "%hu", &(Config::seed_a2));continue;}

		if(!strcmp(p, "PREDICT"))
		{sscanf(v, "%d", &(Config::predict));continue;}

		if(!strcmp(p, "DEBUG_PREDICT"))
		{sscanf(v, "%d", &(Config::debug_predict));continue;}

		if(!strcmp(p, "DEBUG_PREDICT_VALUE"))
		{sscanf(v, "%lf", &(Config::debug_predict_value));continue;}

		if(!strcmp(p, "MAXHISTORY"))
		{sscanf(v, "%d", &(Config::MAXHISTORY));continue;}

		if(!strcmp(p, "SEPARATE_IN_CLUSTERS"))
		{sscanf(v, "%d", &d->SEPARATE_IN_CLUSTERS);continue;}
	}
  
  fclose(fpconfig);
  return(SUCCESS);
}

char *strip_lwsc(char *inp)
{
	char *c, *t;

	// Strip leading Linear White Space Characters 
	c = inp;
	while((*c == ' ')||(*c == '\t')) c++;

	// Strip trailing Linear White Space Characters 
	t = c;
	while(*t) t++;
	t--;
	while((*t == ' ')||(*t == '\t')||(*t == '\n')) *t-- = '\0';

	return(c);
}
