/* -*- Mode:C++ -*- */
/*
 * Copyright (c) Multimedia Communications Lab, Boston University
 *
 * Simulation Program to Evaluate Service Aggregation in VOD
 * =========================================================
 *
 * This work was supported by NSF under grant No. NCR-9523958
 * 
 * DISCLAIMER:
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL
 * MULTIMEDIA COMMUNICATIONS LAB, BOSTON UNIVERSITY OR THE CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 * Contributors:
 *
 * Rajesh Krishnan, 
 * Prithwish Basu,
 * Wang Ke
 */

#include "vod.hpp"

int verbose;
Config config;

// making mvrq global so that predict_arrivals in cluster/rsma.c can call 
MovieRequest mvrq;

// counter to separate client requests into different clusters
// if they arrive in different recomputation clusters
int numRI = 1;

#include "adins.hpp"
#include "channel.hpp"
#include <assert.h>


double Config::debug_at_recomp;
int Config::stt_demand;
int Config::stt_random;
int Config::allow_recomp;
int Config::stt_streams;
double Config::clust_window;
double Config::min_ad_time;
int Config::ratio_max_min_ad;
double Config::max_ad_time;
int Config::ratio_vmin_amin;
double Config::min_video_time;
int Config::ratio_tprime_amax;
double Config::long_term_ad_time;
double Config::long_term_video;
double Config::long_term_window;
int Config::allowing_interactions;
int Config::aggregate;
int Config::plot_time_pos;
double Config::reset_time = -1.0;
int Config::check_eomv = 0;
double Config::stream_id;
double Config::snd_stream;
int Config::join_info;
double Config::user_id;
double Config::wait_distr_power;
double Config::wait_max_time;
double Config::perc_OD;
int Config::max_num_chan;
int Config::return_ondemand;
int Config::recomp_at_event = 1;
CostForm Config::cost_formulation = BIT;
double Config::cost_rate = 0.0;
int Config::cost_mvlen_multiplier = 15;
int Config::show_users_per_stream = 30;
unsigned short int Config::seed_a0 = 27;
unsigned short int Config::seed_a1 = 330;
unsigned short int Config::seed_a2 = 120;
unsigned int Config::predict = 0;
int Config::debug_predict = 0;
double Config::debug_predict_value = 1.0;
unsigned int Config::MAXHISTORY = 100;

// prediction related globals

unsigned int current;
double cum_hist;
double *history;



// Variables to keep track of how many objects are created
// during software execution
int Cost::num_cost = 0;
int Virtualstream::num_vs = 0;
int Channel::num_channel = 0;
int Channel::id_counter = 0;
int Channellist::num_channellist = 0;
int Waitlist::num_waitlist = 0;
int QueueManager::num_queuemanager = 0;
int VSList::num_vslist = 0;
int Ad_Channel::num_ad_channel = 0;
int VSptr::num_vsptr = 0;
int AdStrList::num_adstrlist = 0;
int MStream::num_mstream = 0;
int Cluster::num_cluster = 0;
int Command::num_command = 0;
int PositionNode::num_pn = 0;
template<> int TriangMatrix<int>::num_tmatrix = 0;
template<> int TriangMatrix<double>::num_tmatrix=0;
template<> int TriangMatrix<PositionNode>::num_tmatrix=0;
int Stream::num_stream = 0;
int Clusterer::num_clusterer=0;
int EMCL_Clusterer::num_emcl=0;
int Subscriber::num_subscriber=0;
int Subptr::num_subptr=0;
int Movie::num_mov=0;
int Movielist::num_movlist = 0;
int Node::tot_num_nodes = 0;
int List::tot_num_lists = 0;
int ChannelManager::num_channelmanager = 0;
int WaitStat::num_waitstat = 0;

EMCL_Clusterer ecl;
Ad_Insertion global_ad_ins(&ecl);
extern void print_Stts(Status);
extern void print_Stts(Status, int);

// The function belows keeps track of the number of objects
// created. It can be used to detect memory leak problems.
// Currently unused.
void CheckObjects()
{
  cerr << "CO:  Cost::num_cost = " 
       << Cost::num_cost << endl 
       << "CO:  Virtualstream::num_vs = " 
       << Virtualstream::num_vs << endl 
       << "CO:  Channel::num_channel = " 
       << Channel::num_channel << endl 
       << "CO:  Channellist::num_channellist = " 
       << Channellist::num_channellist << endl 
       << "CO:  ChannelManager::num_channelmanager = " 
       << ChannelManager::num_channelmanager << endl 
       << "CO:  Waitlist::num_waitlist = " 
       << Waitlist::num_waitlist << endl
       << "CO:  QueueManager::num_queuemanager = " 
       << QueueManager::num_queuemanager << endl 
       << "CO:  VSList::num_vslist = " 
       << VSList::num_vslist << endl 
       << "CO:  Ad_Channel::num_ad_channel = " 
       << Ad_Channel::num_ad_channel << endl 
       << "CO:  VSptr::num_vsptr = " 
       << VSptr::num_vsptr << endl 
       << "CO:  AdStrList::num_adstrlist = " 
       << AdStrList::num_adstrlist << endl 
       << "CO:  MStream::num_mstream = " 
       << MStream::num_mstream << endl 
       << "CO:  Cluster::num_cluster = " 
       << Cluster::num_cluster << endl 
       << "CO:  Command::num_command = " 
       << Command::num_command << endl 
       << "CO:  PositionNode::num_pn = " 
       << PositionNode::num_pn << endl 
       << "CO:  TriangMatrix<int>::num_tmatrix = " 
       << TriangMatrix<int>::num_tmatrix << endl 
       << "CO:  TriangMatrix<double>::num_tmatrix = " 
       << TriangMatrix<double>::num_tmatrix << endl 
       << "CO:  TriangMatrix<PositionNode>::num_tmatrix = " 
       << TriangMatrix<PositionNode>::num_tmatrix << endl 
       << "CO:  Stream::num_stream = " 
       << Stream::num_stream << endl 
       << "CO:  Clusterer::num_clusterer = " 
       << Clusterer::num_clusterer << endl 
       << "CO:  EMCL_Clusterer::num_emcl = " 
       << EMCL_Clusterer::num_emcl << endl 
       << "CO:  Subscriber::num_subscriber = " 
       << Subscriber::num_subscriber << endl 
       << "CO:  Subptr::num_subptr = " 
       << Subptr::num_subptr << endl 
       << "CO:  Movie::num_mov = " 
       << Movie::num_mov << endl 
       << "CO:  Movielist::num_movlist = " 
       << Movielist::num_movlist << endl 
       << "CO:  Node::tot_num_nodes = " 
       << Node::tot_num_nodes << endl 
       << "CO:  List::tot_num_lists = " 
       << List::tot_num_lists << endl;
  global_ad_ins.checkTriangMatrix();
  cerr << endl << endl;
}

double IN_PLAYING_STT = -1.0;

// global simulation time
// Making k global. For many functions will need to access its value
double k = 0.0;	

// Creating a pointer to k so that modules that have "k" as variables
// can access the global time variable
double *simulation_time = &k;

// recomp controls whether a recomputation of stream path will take place
// or not.
// set recomp = 1 to force recomputation of stream path according to policy
char recomp = 0;

// newEvent keeps track whether a new arrival or interaction happened
// currently set when there is a new interaction (main.cpp),
// when a stream resumes from interaction (virtstream.C)
// or when the stream reaches end of movie (virtstream.C).

int newEvent = 0;



#define ECHO(a) if(verbose) {cout << (a) << endl;}
#define SHOW(a) if(verbose) {(a); cout << endl;}

#define RQ2ST(rqtype, stat) \
	switch((rqtype)) \
	{ \
		case PAUSE: (stat) = PAUSED; break; \
		case RWIND: (stat) = RWDING; break; \
		case FFWD: (stat) = FWDING; break; \
		case QUIT: (stat) = ATEOMV; break; \
		default: (stat) = UNKNOWN; break; \
	} 


void  GENERATE_VCR_ACTIONS(VCRActionRequest *rq, ChannelManager *cmgr,
			   VSList &rmvstreams, Subscriberlist &activesubpool,
			   AdStrList &adstrlist) 
{
  Status newstat = UNKNOWN; 
  Virtualstream *w = NULL, *v = NULL, *vtmp = NULL; 
  Subptr *s = NULL; 
  Channel *c = NULL;

  // split generate event time and process. Time generation only
  // generates the event time, and the process will be executed when
  // GENERATE_VCR_ACTIONS is called. The purpose of this separation
  // is so that it will only generate the requesters when it is truly
  // going to be called. 

  rq->GenerateEventProcess();

  if(rq->Requester()) {
    RQ2ST(rq->Requester()->GetRequest(),newstat); 
  }
  else {
    // in this case no suitable requester has been found
    // so no VCR action can be done at the current time.
    // A new interaction time must be set
    rq->GenerateEventTime();
    return;     
  }

  if(newstat != UNKNOWN) {
    w = rq->Requester()->GetMystream(); 
    double x = k + rq->Requester()->GetDuration(); 

    if (w->Getstatus() != AD_INS) { 
      if(w->NumUsers() == 1) { 
	assert(w->GetChan());
	cmgr->returnActiveChannel(w->GetChan());
	w->SetChan(NULL);

	if ( (config.data.AGG_POLICY == ADINSPOL) ) {
	  if (w->been_interaction == 0) {
	    v = w->GenerateBogus();
	    v->SwitchId(w);
	    
	    if (config.data.AGG_POLICY == ADINSPOL) 
	      v->to_merge_with = global_ad_ins.to_merge_with(v->StreamId());
	    else 
	      v->to_merge_with = -1.0;
	    
	    assert(v->bogus);
	    v->SetChan(NULL);
	    w->SetChan(NULL);
	    rmvstreams.Prepend(v,NULL);
	  }
	  w->been_interaction = 1;
	}

	w->Setstatus(newstat); 
	// set until to x-1 because the inquiry is made at the beginning
	// of the time second frame. Thus, if a request is serviced at
	// time 30 with duration 5, it should end after time 34
	w->Setuntil(x-1); 

	if (cmgr->contingencyAvailable()) {
	  c = cmgr->getIdleChannel(Cont);
	  if (!c) {
	    cerr << "GENERATE_VCR_ACTIONS: ERROR: Contingency available "
		 << "yet returning null channel vector\n";
	    assert(c);
	  }
	  
	  w->SetChan(c);
	  c->setVS(w);
	  cmgr->putToActive(c);
	  assert( (c->getType() == Cont) );

	}
	else {
	  // the VS must be detached for it is waiting for a channel
	  // and should not be allowed to progress in the movie stream
	  vtmp = (Virtualstream *)rmvstreams.Detach(w);
	  assert((vtmp == w));
	  cmgr->waitForContingencyChannel(w);
	}
      }    
      else { 
	
	v = w->Forkstream(0, newstat, (x-1)); 
	s = w->Findsubptr(rq->Requester()); 
	assert(v);
	assert(s);
	if (s == NULL) {
	  cerr << "ERROR: GEN_VCR_ACT: checking requester:\n";
	  rq->Requester()->Dumpdata();
	  cerr << "ERROR: GEN_VCR_ACT: checking stream:\n";
	  w->Dumpdata();
	  assert(s);
	}
	v->Adduser(w->Detachuser(s)); 
	
	if ( (config.data.AGG_POLICY == ADINSPOL) )
	  v->been_interaction = 1;
	
	if (cmgr->contingencyAvailable()) {
	  c = cmgr->getIdleChannel(Cont);
	  if (!c) {
	    cerr << "GENERATE_VCR_ACTIONS: ERROR: contingency available but"
		 << " returning null channel pointer\n";
	    assert(c);
	  }
	  v->SetChan(c);
	  c->setVS(v);
	  cmgr->putToActive(c);
	  rmvstreams.Prepend(v, NULL); 
	}
	else {
	  cmgr->waitForContingencyChannel(v);
	}
      }
      
      // New interaction new event in the system
      newEvent = 1;
    }
    else { 

      assert( (config.data.AGG_POLICY == ADINSPOL) );
      Subscriber *subs = (Subscriber *)activesubpool.Detach(rq->Requester());
      adstrlist.Store(subs,newstat,(x-k)); 
    } 
    rq->Requester()->Passivate(); 

  }
  
  rq->GenerateEventTime();
}

void BATCH_REQUESTS(MovieRequest *mvrq, Subscriberlist *activesubpool, 
		    VSList *bmvstreams, ChannelManager *chanmgr)  
{
  while (k >= mvrq->EventTime()) {

    if(mvrq->Requester()) {
      
      Virtualstream *ww, *vv; 
      cout << "\n\nMovie Request at " << k << endl; 
      
      vv = new Virtualstream(0, mvrq->Mvchoice()); 

      if (config.data.SEPARATE_IN_CLUSTERS) {
	Movie *tmv = vv->GetMv();
	tmv->SetMovieId(numRI);
      }

      vv->newArrival = 1;
      vv->Adduser(new Subptr(mvrq->Requester())); 
      activesubpool->Prepend(mvrq->Requester(), NULL); 
      
      if((ww=bmvstreams->GetMatch(vv))) {
	cout << "match successful\n\n"; 
	assert(ww->GetChan());
	assert( (ww->StreamId() != vv->StreamId() ) );

	ww->Joinstream(vv, bmvstreams, chanmgr); 

	// check to see how this newArrival variable should be handled
	// when the batching interval is different from 0
	ww->newArrival = 1;

	ww->n_batched += 1;
      }
      else {

	// The line belows allows the streams to be batched according to 
	// config.data.BATCH_INTERVAL value. 

	vv->Setuntil(k+config.data.BATCH_INTERVAL);

	if ((chanmgr->onDemandAvailable() > 0)) {
	  Channel *c = chanmgr->getIdleChannel(OnDem);
	  if (c) {
	    vv->SetChan(c);
	    c->setVS(vv);
	  }
	  else {
	    cerr << "ERROR: BATCH_REQUESTS: onDemandAvailable > 0 but channel NULL\n\n";
	    assert(c);
	  }
	  cout << "match failed\n\n";
	  chanmgr->putToActive(c);
	  bmvstreams->Prepend(vv, NULL);  
	}
	else {
	  chanmgr->waitForOnDemandChannel(vv);
	}
      }
    }
    else {
      cout << "WARNING: BATCH_REQUESTS called without a valid requester. Attempting new movie request\n";
    }
    mvrq->GenerateEvent(); 
  }
}
 
int main(int argc, char **argv)
{
  extern double NetBandwidth;
  int tick = 0;
  double window = 0.0;
  char interact_recomp = 0;		
  unsigned short int a[3];
  
  if(argc !=2)
    {
      cout << "Usage: " << argv[0] << " <configfile>" << endl;
      exit(0);
    }

  ECHO("Reading Configuration File...");
  if(config.Setup(argv[1]) == FAILURE)
    {
      cout << "Error in configuration" << endl;
      exit(0);
    }

  verbose = config.data.VERBOSE;
  SHOW(config.Dumpdata());


  // Set the seed for the random number generator
  ECHO("Setting seed for Random Number Generator...");
  // using seed48 and lrand48 to generate random numbers
  a[0] = Config::seed_a0;
  a[1] = Config::seed_a1;
  a[2] = Config::seed_a2;

  seed48(a);

  // initialize prediction related globals
  current = 0;
  cum_hist = 0;
  history = (double *) calloc(Config::MAXHISTORY, sizeof(double));
  

  // Initialize the movie database
  ECHO("Initializing Movie Database...");
  Movielist mvpool(config.data.MOVIE_DB);
  ECHO("Zipf Distributing Movie Database...");
  mvpool.Zipf_distribute(config.data.ZIPF_PARAM);

  // Initialize the subscriber database
  ECHO("Initializing Subscriber Database...");
  Subscriberlist idlesubpool(config.data.TOT_USERS);
  Subscriberlist activesubpool(0);

  // Declare variables for the simulation
  VSList bmvstreams;    // Batched Movie Streams
  VSList rmvstreams;	// Running Movie Streams
  VSList vmvstreams;	// Vacated Movie Streams: not implemented yet

  ChannelManager chanmgr(Config::max_num_chan, Config::perc_OD);
  AdStrList adstrlist;  // List of streams in ad_ins that requested interaction

  Ad_Channel adchannel(AMIN);
  adchannel.set_last_ad_time(0.0);


  // Initialize the request and interaction generators
  ECHO("Initializing Event Generators...");
  mvrq.ResetParameters(config.data.MV_RQ_RATE, &mvpool, &idlesubpool);
  VCRActionRequest quitrq(config.data.QT_RQ_RATE, 0,QUIT, &activesubpool);
  VCRActionRequest ffwdrq(config.data.FF_RQ_RATE, config.data.FF_DUR,FFWD, &activesubpool);
  VCRActionRequest rwndrq(config.data.RW_RQ_RATE, config.data.RW_DUR,RWIND, &activesubpool);
  VCRActionRequest pausrq(config.data.PAU_RQ_RATE, config.data.PAU_DUR,PAUSE, &activesubpool);

  // Run the Simulation
  ECHO("Starting the Simulation...");

  int j = 0;
  double last_computed = 0.0;
  double reset_time = 0.0;
  char in_reset = 0;
  
  while(j < config.data.MAXSIMCLOCK)
    {
      cout << "STATS: T: "<< k << " S: "<<rmvstreams.GetNumberVS()
	   << " U: " 
	   << (activesubpool.Numnodes() + adstrlist.Numnodes())
	   << "\n";
    
      if ( Config::show_users_per_stream > 0) {
	if ( modulus(tick,Config::show_users_per_stream) == 0) {
	  cout << "STR: " << k << " : " ; rmvstreams.UsersPerStream(); cout << endl;
	}
      }
    
      // return those impatient requesters to idlesubpool
      chanmgr.checkWaitUntilLimits(&activesubpool, &idlesubpool);

      // allocate available channels to those waiting in queue
      chanmgr.serveOnDemandRequests(&bmvstreams);
      chanmgr.serveContingencyRequests(&rmvstreams);
    
      assert( (chanmgr.qmgr->onDemandlist.wstat.stream_waiting == \
	       chanmgr.qmgr->onDemandlist.Numnodes()) );
      assert( (chanmgr.qmgr->contingencylist.wstat.stream_waiting == \
	       chanmgr.qmgr->contingencylist.Numnodes()) );

      // Generate Movie Requests and Batch 
      BATCH_REQUESTS(&mvrq, &activesubpool, &bmvstreams, &chanmgr);
    
      // Generate interactions and fork streams here
      VCRActionRequest *va[4], *tmpva;

      // At 30s (AMIN) intervals for ADINSPOL
      if ( ( (config.data.AGG_POLICY != ADINSPOL) || \
	     ( (config.data.AGG_POLICY == ADINSPOL) && \
	       (fabs(adchannel.get_nearest_ad_time(k)-k) < EPSILON)
	       ) 
	     ) ) {
      
	// If we are having interactions
	if (Config::allowing_interactions) {
	
	  // variables to hold the pointers to the proper requests
	  va[0] = &quitrq; va[1] = &ffwdrq;
	  va[2] = &rwndrq; va[3] = &pausrq; 
	  tmpva = NULL;
	
	  for( ; ;) {
	    // bubble sort all requests stored in the 4 request lists
	    for (int bubblesort = 0; bubblesort < 4; bubblesort++) {
	      for (int vaind = 0; vaind < 3; vaind++) {
		if (va[vaind]->EventTime() > va[vaind+1]->EventTime()) {
		  tmpva = va[vaind+1];
		  va[vaind+1] = va[vaind];
		  va[vaind] = tmpva;
		}
	      }
	    }
	    // at this point va should contain in increasing event-time
	    // order the requests
	    for (int assertion=0; assertion < 3; assertion++) {
	      assert((va[assertion]->EventTime() <= va[assertion+1]->EventTime()));
	    }
	  
	    // still not time for next event
	    if (va[0]->EventTime() > k) {
	      break;
	    }
	  
	    // it's time for event, so generate the vcr action. This will be repeated
	    // until a event time becomes greater than the current time and the loop
	    // is broken by the condition above
	    GENERATE_VCR_ACTIONS(va[0], &chanmgr,rmvstreams,activesubpool,adstrlist);
	  }
	}
      
	// if there are nodes stored at the ad stream list, dispatch them
	if (adstrlist.Numnodes() > 0) {
	  assert( (config.data.AGG_POLICY == ADINSPOL) );
	  adstrlist.Dispatch(&chanmgr, &activesubpool, &rmvstreams, k);
	}
      
	// Advance Batched Streams only 
	if (Config::plot_time_pos)
	  cout << "launching batch streams - ";
      
	bmvstreams.Advanceall(k);

	if (Config::plot_time_pos)
	  cout << endl;
      
	// Advanceall() sets batch ready to play on timeout by default.
	// Other policies must set RTOPLAY appropriately.
	// Optimal MFQL policy recommended -- Aggarwal et al.
      
      
	// Transfer ready-to-play users to running streams list
	rmvstreams.Launchbatch(&bmvstreams, k);
      
      }
    
      // Set DSAP recompute flag here based on triggering policy
      recomp = 0;
    
      // if k-last_computed >= CLUST_RECOMP, then it is time
      // for the streams to enter the reset ad interval.
      // This was suggested for ad insertion. Currently unused

      if ( (k - last_computed >= config.data.CLUST_RECOMP) && (!in_reset)) {
      
	interact_recomp = 0;
      
	in_reset = 1;

	if (config.data.SEPARATE_IN_CLUSTERS) {
	  numRI++;
	}
      
	// set Config::reset_time = 0 through CONFIG.adins to avoid
	// spending time in the "reset ad statistic interval". See CONFIG.adins
	// and adins.hpp
	reset_time = (Config::reset_time < 0) ?   \
	  ceil((LAST_VIDEO_TIME)/(AMAX))*(AMAX) : \
	  Config::reset_time;
      
	if (reset_time > 0) 
	  rmvstreams.ResetAdTime();
      
	if (fabs(adchannel.get_nearest_ad_time(k)-k) > EPSILON) {
	  cerr << "ERROR: trying to execute recomputation out of parameter\n";
	  exit(1);
	}
      }
    
      // Apply suitable DSAP Algorithm, recomputing if necessary
#ifdef INITIALGREEDY    
      if (k == 0.0) {
	window = config.data.TOT_USERS/config.data.MV_RQ_RATE;
	rmvstreams.Aggregate(2, recomp, &window, k);
      }
      else if (k < (config.data.TOT_USERS/config.data.MV_RQ_RATE)) {
	if (window <= EPSILON) {
	  recomp = 1;
	  window = config.data.CLUST_RECOMP; 
	  rmvstreams.Aggregate(config.data.AGG_POLICY, recomp, &window, k);
	  last_computed = k + config.data.SIMGRAIN;
	}
	else rmvstreams.Aggregate(2, recomp, &window, k);
      }
      else {
	if (window <= EPSILON)
	  window = config.data.CLUST_RECOMP;
	rmvstreams.Aggregate(config.data.AGG_POLICY, recomp, &window, k); 
      }

#else
      if (window <= EPSILON)
	window = config.data.CLUST_RECOMP;
#endif


      // if the streams are in the reset interval but the
      // time has reached zero, then it is time for recomp
      // and exiting reset interval. The current time k is
      // then the time recomp is called.

      if ((reset_time <= 0.0) && (in_reset)) {
	cout << "Setting recomp at " << k << endl;
	recomp = 1;
	in_reset = 0;
	last_computed = k;
      }

      // if we are doing recomputation at every event 
      if (Config::recomp_at_event) {
	recomp = 0; // do not recompute unless there is a new event

	if (newEvent) {
	  recomp = 1;
	  in_reset = 0; // assure we will not be in the in_reset interval
	  last_computed = k; // put last_computed as k so that the CLUST_RECOMP 
				// interval will not trigger a recomputation
	  newEvent = 0;
	}
      }	


      // only call aggregate methods if the streams are not
      // in the reset interval.

      if (!in_reset) {
	if (Config::aggregate) {

	  // RemoveBogus removes "bogus" streams. These "bogus" streams
	  // are "virtual" streams created to keep ad-insertion algorithms
	  // working properly.
	  if (recomp) 
	    rmvstreams.RemoveBogus();
	
	  rmvstreams.Aggregate(config.data.AGG_POLICY, recomp,\
			       &(Config::clust_window), k);    
	  
	}
      }
    
      // Advance Running Streams -    
      rmvstreams.Advanceall(k);
    
      // Handle Exits (and Quits) Here
      rmvstreams.Handleexits(&activesubpool, &idlesubpool, &chanmgr);
    
      // Find mergable streams in the set of running streams 
      // and remove trailing streams
      rmvstreams.Compact(&chanmgr);


      // Advance Simulation Clock -
      // The simulation clock is advanced before the aggregate function is called
      // because for the kth second behavior, the inquiry should be made at
      // time k-1. i.e., if I want to know the behavior of the stream on its 100th
      // second, I should inquire at time == 99 seconds.
      j++;
      k += config.data.SIMGRAIN;
      tick++;
      reset_time -= config.data.SIMGRAIN;
 
      if (j%1800 == 0)
	cout  << j/30 << "\t" << config.data.TOT_USERS << "\t"	
	      << activesubpool.Numnodes() << "\t" 
	      << rmvstreams.Numnodes() << endl; 
    
    }
  
  ECHO("Statistics at completion........");
  ECHO("Total number of users in the system :");
  ECHO(activesubpool.Numnodes());
  ECHO("Total number of streams in the system :");
  ECHO(rmvstreams.Numnodes());

  if ( (config.data.AGG_POLICY == ADINSPOL) ) {
    cout << "\nAd_Insertion bandwidth characteristics:\n";
    rmvstreams.GetCostBandwidth();
    cout << endl;
  }
  
  cout << "AVG_WAIT_RECOMP: "<< k << " = " << rmvstreams.totalTTR 
       << "/" << rmvstreams.nStreamsTTR;
  if (rmvstreams.nStreamsTTR) 
    cout << " = " << (rmvstreams.totalTTR/rmvstreams.nStreamsTTR);
  cout << "\n\n";
  
  double TotalBandwidth = (config.data.MAXSIMCLOCK*config.data.SIMGRAIN*activesubpool.Numnodes()*FRAME_RATE);
  

  cout << j/30 << "\t"
       << config.data.TOT_USERS << "\t"	
       << activesubpool.Numnodes() << "\t" 
       << rmvstreams.Numnodes() << endl;

  SHOW(cout << config.data.MV_RQ_RATE << "\t"
       << config.data.FF_RQ_RATE << "\t"
       << config.data.CLUST_RECOMP << "\t"
       << TotalBandwidth - NetBandwidth << "\t"
       << ((TotalBandwidth - NetBandwidth)/TotalBandwidth)*100 << endl);
       

  ECHO("Simulation Completed Successfully.");
}

// CAVEATS :
// Users forwarding past the end-of-movie will leave.
// Simultaneous movie requests are merged automatically.
// VCR Actions possible only if current request state is PLAY.
// Currently the only batching policy supported is timeout based.
// All streams currently run on a dummy channel. 
// There is no limit on the number of channels now. 
// Need to write GetChannel(), Block() and Vacate()
// Users reneging from batch not yet implemented.
// User penalty not yet implemented.
// Vacations not yet implemented.

