/* -*- Mode:C++ -*- */
/*
 * Copyright (c) Multimedia Communications Lab, Boston University
 *
 * Simulation Program to Evaluate Service Aggregation in VOD
 * =========================================================
 *
 * This work was supported by NSF under grant No. NCR-9523958
 * 
 * DISCLAIMER:
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL
 * MULTIMEDIA COMMUNICATIONS LAB, BOSTON UNIVERSITY OR THE CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 * Contributors:
 *
 * Rajesh Krishnan
 */

#include "vod.hpp"

void Movie::Dumpdata(void)
{
	cout << "MV_ID = " << movie_id << "\t";
	cout << "FREQ = " << frequency << "\t";
	cout << "DUR = " << duration << "\n";
}

Movielist::Movielist(char *filename)
{
	int freq, id;
	double dur;
	ifstream dbfile;
	char junk[132];

	dbfile.open(filename);
	if(!dbfile)
	{
		cout << "Error Opening " << filename << "\n";
		exit(1);
	}

	while(1)
	{
		dbfile >> id;
		dbfile >> dur;
		dur *= 60; 	// Convert to seconds
		dbfile >> freq;
		dbfile.getline(junk, 131);
		if(dbfile.eof()) break;
		Prepend(new Movie(dur, freq, id), NULL);
	}

	dbfile.close();
	Computethresholds();
	num_movlist++;
}
	
void Movielist::Computethresholds(void)
{
	Movie *cursor = (Movie *)head;
	int cum_freq = 0;
	while(cursor)
	{
		cum_freq += cursor->frequency;
		cursor->threshold = cum_freq;
		cursor = (Movie *)cursor->next;
	}
}

Movie *Movielist::Randomchoice(void)
{
	long outcome =  1 + random() % ((Movie *)tail)->threshold;

	Movie *cursor = (Movie *)head;
	while(cursor)
	{
		if(outcome <= cursor->threshold) return((Movie *)cursor);
		cursor = (Movie *)cursor->next;
	}

	return(NULL);
}

void Movielist::Zipf_distribute(double param)
{
	Movie *cursor = (Movie *)head;
	double f;

	// Zipf distribution : f(i) = c/i^(1-theta)
	// c = BIASPRECISION = Normalization Constant

	while(cursor)
	{
		f = pow((double)cursor->movie_id, 1 - param);
		f = BIASPRECISION/f;
		cursor->frequency = (int) ceil(f);
		cursor = (Movie *)cursor->next;
	}
	Computethresholds();
}


