/* -*- Mode:C++ -*- */
/*
 * Copyright (c) Multimedia Communications Lab, Boston University
 *
 * Simulation Program to Evaluate Service Aggregation in VOD
 * =========================================================
 *
 * This work was supported by NSF under grant No. NCR-9523958
 * 
 * DISCLAIMER:
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL
 * MULTIMEDIA COMMUNICATIONS LAB, BOSTON UNIVERSITY OR THE CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 * Contributors:
 *
 * Rajesh Krishnan, 
 * Prithwish Basu,
 * Wang Ke
 */

#include "vod.hpp"

#include <stdlib.h>

#include "config.hpp"
#include "adins.hpp"
#include "virtstream.hpp"
#include <assert.h>


extern double k; // decleared in main.cpp
extern int verbose ; // declared in main.cpp

double waitDistribution(double x01)
{
  double y = 1.0 /Config::wait_distr_power;
  return (pow(x01, y));
}

Subscriber::~Subscriber()
{
  num_subscriber--;
}
void Subscriber::Dumpdata()
{  
  cout << "User_Id " << user_id << " Request ";
  cout << (int)request << " ";VCRActionRequest::pType(request);
  cout << " Duration " << duration << "\n";
  cout << "virt: " << ((mystream == NULL) ? -1: mystream->StreamId()) << endl;
  cout << " start_wait_time:" << start_wait_time << " - end_wait_time:"
       << end_wait_time << " - wait_until:" << wait_until
       << " - duration(w_u-s_w_t):" << (wait_until - start_wait_time) << endl;
}

void Subscriber::SetMystream(Virtualstream *s)
{
  mystream = s;
}

void Subscriber::SetWaitUntil()
{
  wait_until = start_wait_time + (Config::wait_max_time *
				  pow( ((1.0*random())/(1.0 * RAND_MAX)), 
				       (1.0/Config::wait_distr_power)));  
  if (verbose) {
    cout << "Subscriber " << user_id << " waiting from "
	 << start_wait_time << " until " << wait_until << " ("
	 << (wait_until-start_wait_time) << ")" << endl;
  }
  assert(wait_until >= start_wait_time);
}

Subscriberlist::Subscriberlist(int n)
{
	while(n > 0) 
	{
		Prepend(new Subscriber(n), NULL);
		n--;
	}
}

Subscriber *Subscriberlist::ExtractRandom(void)
{
	if(!num_nodes) return(NULL);
	int outcome =  (int) random() % num_nodes;
	return((Subscriber *)Detach(Seek(head, outcome, FORWARD)));	
}

Subscriber *Subscriberlist::RandomChoice(void)
{
	if(!num_nodes) return(NULL);
	int outcome =  (int) random() % num_nodes;
	return ((Subscriber *)Seek(head, outcome, FORWARD));
}

Subptr::~Subptr()
{
  num_subptr--;
  if (sptr)
    if (sptr->GetUID() == Config::user_id) {
      cout << "User " << sptr->GetUID() << " deleted\n";
    }
}





